# This file was auto-generated by Fern from our API Definition.

import typing
from json.decoder import JSONDecodeError

from ..core.api_error import ApiError
from ..core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ..core.http_response import AsyncHttpResponse, HttpResponse
from ..core.jsonable_encoder import jsonable_encoder
from ..core.pydantic_utilities import parse_obj_as
from ..core.request_options import RequestOptions
from ..core.serialization import convert_and_respect_annotation_metadata
from ..errors.bad_request_error import BadRequestError
from ..errors.conflict_error import ConflictError
from ..errors.forbidden_error import ForbiddenError
from ..errors.internal_server_error import InternalServerError
from ..errors.not_found_error import NotFoundError
from ..errors.service_unavailable_error import ServiceUnavailableError
from ..errors.too_many_requests_error import TooManyRequestsError
from ..errors.unauthorized_error import UnauthorizedError
from ..types.not_found_error_body import NotFoundErrorBody
from ..types.permission_denied_error import PermissionDeniedError
from ..types.rate_limit_error import RateLimitError
from ..types.unauthorized_error_body import UnauthorizedErrorBody
from .types.functions_create_request_dependencies_value import (
    FunctionsCreateRequestDependenciesValue,
)
from .types.functions_create_response import FunctionsCreateResponse
from .types.functions_find_by_hash_response import FunctionsFindByHashResponse
from .types.functions_get_by_env_response import FunctionsGetByEnvResponse
from .types.functions_get_response import FunctionsGetResponse
from .types.functions_list_by_env_response import FunctionsListByEnvResponse
from .types.functions_list_response import FunctionsListResponse

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class RawFunctionsClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._client_wrapper = client_wrapper

    def list(
        self, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[FunctionsListResponse]:
        """
        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[FunctionsListResponse]
            Success
        """
        _response = self._client_wrapper.httpx_client.request(
            "functions",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    FunctionsListResponse,
                    parse_obj_as(
                        type_=FunctionsListResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        UnauthorizedErrorBody,
                        parse_obj_as(
                            type_=UnauthorizedErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        PermissionDeniedError,
                        parse_obj_as(
                            type_=PermissionDeniedError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        NotFoundErrorBody,
                        parse_obj_as(
                            type_=NotFoundErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 429:
                raise TooManyRequestsError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        RateLimitError,
                        parse_obj_as(
                            type_=RateLimitError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 503:
                raise ServiceUnavailableError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(
                status_code=_response.status_code,
                headers=dict(_response.headers),
                body=_response.text,
            )
        raise ApiError(
            status_code=_response.status_code,
            headers=dict(_response.headers),
            body=_response_json,
        )

    def create(
        self,
        *,
        code: str,
        hash: str,
        signature: str,
        signature_hash: str,
        name: str,
        language: str,
        description: typing.Optional[str] = OMIT,
        tags: typing.Optional[typing.Sequence[str]] = OMIT,
        metadata: typing.Optional[typing.Dict[str, typing.Optional[str]]] = OMIT,
        dependencies: typing.Optional[
            typing.Dict[str, typing.Optional[FunctionsCreateRequestDependenciesValue]]
        ] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[FunctionsCreateResponse]:
        """
        Parameters
        ----------
        code : str

        hash : str

        signature : str

        signature_hash : str

        name : str

        language : str

        description : typing.Optional[str]

        tags : typing.Optional[typing.Sequence[str]]

        metadata : typing.Optional[typing.Dict[str, typing.Optional[str]]]

        dependencies : typing.Optional[typing.Dict[str, typing.Optional[FunctionsCreateRequestDependenciesValue]]]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[FunctionsCreateResponse]
            Success
        """
        _response = self._client_wrapper.httpx_client.request(
            "functions",
            method="POST",
            json={
                "code": code,
                "hash": hash,
                "signature": signature,
                "signatureHash": signature_hash,
                "name": name,
                "language": language,
                "description": description,
                "tags": tags,
                "metadata": metadata,
                "dependencies": convert_and_respect_annotation_metadata(
                    object_=dependencies,
                    annotation=typing.Optional[
                        typing.Dict[
                            str,
                            typing.Optional[FunctionsCreateRequestDependenciesValue],
                        ]
                    ],
                    direction="write",
                ),
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    FunctionsCreateResponse,
                    parse_obj_as(
                        type_=FunctionsCreateResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        UnauthorizedErrorBody,
                        parse_obj_as(
                            type_=UnauthorizedErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        PermissionDeniedError,
                        parse_obj_as(
                            type_=PermissionDeniedError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        NotFoundErrorBody,
                        parse_obj_as(
                            type_=NotFoundErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 409:
                raise ConflictError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 429:
                raise TooManyRequestsError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        RateLimitError,
                        parse_obj_as(
                            type_=RateLimitError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 503:
                raise ServiceUnavailableError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(
                status_code=_response.status_code,
                headers=dict(_response.headers),
                body=_response.text,
            )
        raise ApiError(
            status_code=_response.status_code,
            headers=dict(_response.headers),
            body=_response_json,
        )

    def get(
        self, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[FunctionsGetResponse]:
        """
        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[FunctionsGetResponse]
            Success
        """
        _response = self._client_wrapper.httpx_client.request(
            f"functions/{jsonable_encoder(id)}",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    FunctionsGetResponse,
                    parse_obj_as(
                        type_=FunctionsGetResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        UnauthorizedErrorBody,
                        parse_obj_as(
                            type_=UnauthorizedErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        PermissionDeniedError,
                        parse_obj_as(
                            type_=PermissionDeniedError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        NotFoundErrorBody,
                        parse_obj_as(
                            type_=NotFoundErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 429:
                raise TooManyRequestsError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        RateLimitError,
                        parse_obj_as(
                            type_=RateLimitError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 503:
                raise ServiceUnavailableError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(
                status_code=_response.status_code,
                headers=dict(_response.headers),
                body=_response.text,
            )
        raise ApiError(
            status_code=_response.status_code,
            headers=dict(_response.headers),
            body=_response_json,
        )

    def delete(
        self, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[None]:
        """
        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[None]
        """
        _response = self._client_wrapper.httpx_client.request(
            f"functions/{jsonable_encoder(id)}",
            method="DELETE",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                return HttpResponse(response=_response, data=None)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        UnauthorizedErrorBody,
                        parse_obj_as(
                            type_=UnauthorizedErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        PermissionDeniedError,
                        parse_obj_as(
                            type_=PermissionDeniedError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        NotFoundErrorBody,
                        parse_obj_as(
                            type_=NotFoundErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 429:
                raise TooManyRequestsError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        RateLimitError,
                        parse_obj_as(
                            type_=RateLimitError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 503:
                raise ServiceUnavailableError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(
                status_code=_response.status_code,
                headers=dict(_response.headers),
                body=_response.text,
            )
        raise ApiError(
            status_code=_response.status_code,
            headers=dict(_response.headers),
            body=_response_json,
        )

    def findbyhash(
        self, hash: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[FunctionsFindByHashResponse]:
        """
        Parameters
        ----------
        hash : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[FunctionsFindByHashResponse]
            Success
        """
        _response = self._client_wrapper.httpx_client.request(
            f"functions/hash/{jsonable_encoder(hash)}",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    FunctionsFindByHashResponse,
                    parse_obj_as(
                        type_=FunctionsFindByHashResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        UnauthorizedErrorBody,
                        parse_obj_as(
                            type_=UnauthorizedErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        PermissionDeniedError,
                        parse_obj_as(
                            type_=PermissionDeniedError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        NotFoundErrorBody,
                        parse_obj_as(
                            type_=NotFoundErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 429:
                raise TooManyRequestsError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        RateLimitError,
                        parse_obj_as(
                            type_=RateLimitError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 503:
                raise ServiceUnavailableError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(
                status_code=_response.status_code,
                headers=dict(_response.headers),
                body=_response.text,
            )
        raise ApiError(
            status_code=_response.status_code,
            headers=dict(_response.headers),
            body=_response_json,
        )

    def getbyenv(
        self,
        organization_id: str,
        project_id: str,
        environment_id: str,
        function_id: str,
        *,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[FunctionsGetByEnvResponse]:
        """
        Parameters
        ----------
        organization_id : str

        project_id : str

        environment_id : str

        function_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[FunctionsGetByEnvResponse]
            Success
        """
        _response = self._client_wrapper.httpx_client.request(
            f"organizations/{jsonable_encoder(organization_id)}/projects/{jsonable_encoder(project_id)}/environments/{jsonable_encoder(environment_id)}/functions/{jsonable_encoder(function_id)}",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    FunctionsGetByEnvResponse,
                    parse_obj_as(
                        type_=FunctionsGetByEnvResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        UnauthorizedErrorBody,
                        parse_obj_as(
                            type_=UnauthorizedErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        PermissionDeniedError,
                        parse_obj_as(
                            type_=PermissionDeniedError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        NotFoundErrorBody,
                        parse_obj_as(
                            type_=NotFoundErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 429:
                raise TooManyRequestsError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        RateLimitError,
                        parse_obj_as(
                            type_=RateLimitError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 503:
                raise ServiceUnavailableError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(
                status_code=_response.status_code,
                headers=dict(_response.headers),
                body=_response.text,
            )
        raise ApiError(
            status_code=_response.status_code,
            headers=dict(_response.headers),
            body=_response_json,
        )

    def listbyenv(
        self,
        organization_id: str,
        project_id: str,
        environment_id: str,
        *,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[FunctionsListByEnvResponse]:
        """
        Parameters
        ----------
        organization_id : str

        project_id : str

        environment_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[FunctionsListByEnvResponse]
            Success
        """
        _response = self._client_wrapper.httpx_client.request(
            f"organizations/{jsonable_encoder(organization_id)}/projects/{jsonable_encoder(project_id)}/environments/{jsonable_encoder(environment_id)}/functions",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    FunctionsListByEnvResponse,
                    parse_obj_as(
                        type_=FunctionsListByEnvResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        UnauthorizedErrorBody,
                        parse_obj_as(
                            type_=UnauthorizedErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        PermissionDeniedError,
                        parse_obj_as(
                            type_=PermissionDeniedError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        NotFoundErrorBody,
                        parse_obj_as(
                            type_=NotFoundErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 429:
                raise TooManyRequestsError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        RateLimitError,
                        parse_obj_as(
                            type_=RateLimitError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 503:
                raise ServiceUnavailableError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(
                status_code=_response.status_code,
                headers=dict(_response.headers),
                body=_response.text,
            )
        raise ApiError(
            status_code=_response.status_code,
            headers=dict(_response.headers),
            body=_response_json,
        )


class AsyncRawFunctionsClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._client_wrapper = client_wrapper

    async def list(
        self, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[FunctionsListResponse]:
        """
        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[FunctionsListResponse]
            Success
        """
        _response = await self._client_wrapper.httpx_client.request(
            "functions",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    FunctionsListResponse,
                    parse_obj_as(
                        type_=FunctionsListResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        UnauthorizedErrorBody,
                        parse_obj_as(
                            type_=UnauthorizedErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        PermissionDeniedError,
                        parse_obj_as(
                            type_=PermissionDeniedError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        NotFoundErrorBody,
                        parse_obj_as(
                            type_=NotFoundErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 429:
                raise TooManyRequestsError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        RateLimitError,
                        parse_obj_as(
                            type_=RateLimitError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 503:
                raise ServiceUnavailableError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(
                status_code=_response.status_code,
                headers=dict(_response.headers),
                body=_response.text,
            )
        raise ApiError(
            status_code=_response.status_code,
            headers=dict(_response.headers),
            body=_response_json,
        )

    async def create(
        self,
        *,
        code: str,
        hash: str,
        signature: str,
        signature_hash: str,
        name: str,
        language: str,
        description: typing.Optional[str] = OMIT,
        tags: typing.Optional[typing.Sequence[str]] = OMIT,
        metadata: typing.Optional[typing.Dict[str, typing.Optional[str]]] = OMIT,
        dependencies: typing.Optional[
            typing.Dict[str, typing.Optional[FunctionsCreateRequestDependenciesValue]]
        ] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[FunctionsCreateResponse]:
        """
        Parameters
        ----------
        code : str

        hash : str

        signature : str

        signature_hash : str

        name : str

        language : str

        description : typing.Optional[str]

        tags : typing.Optional[typing.Sequence[str]]

        metadata : typing.Optional[typing.Dict[str, typing.Optional[str]]]

        dependencies : typing.Optional[typing.Dict[str, typing.Optional[FunctionsCreateRequestDependenciesValue]]]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[FunctionsCreateResponse]
            Success
        """
        _response = await self._client_wrapper.httpx_client.request(
            "functions",
            method="POST",
            json={
                "code": code,
                "hash": hash,
                "signature": signature,
                "signatureHash": signature_hash,
                "name": name,
                "language": language,
                "description": description,
                "tags": tags,
                "metadata": metadata,
                "dependencies": convert_and_respect_annotation_metadata(
                    object_=dependencies,
                    annotation=typing.Optional[
                        typing.Dict[
                            str,
                            typing.Optional[FunctionsCreateRequestDependenciesValue],
                        ]
                    ],
                    direction="write",
                ),
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    FunctionsCreateResponse,
                    parse_obj_as(
                        type_=FunctionsCreateResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        UnauthorizedErrorBody,
                        parse_obj_as(
                            type_=UnauthorizedErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        PermissionDeniedError,
                        parse_obj_as(
                            type_=PermissionDeniedError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        NotFoundErrorBody,
                        parse_obj_as(
                            type_=NotFoundErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 409:
                raise ConflictError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 429:
                raise TooManyRequestsError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        RateLimitError,
                        parse_obj_as(
                            type_=RateLimitError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 503:
                raise ServiceUnavailableError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(
                status_code=_response.status_code,
                headers=dict(_response.headers),
                body=_response.text,
            )
        raise ApiError(
            status_code=_response.status_code,
            headers=dict(_response.headers),
            body=_response_json,
        )

    async def get(
        self, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[FunctionsGetResponse]:
        """
        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[FunctionsGetResponse]
            Success
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"functions/{jsonable_encoder(id)}",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    FunctionsGetResponse,
                    parse_obj_as(
                        type_=FunctionsGetResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        UnauthorizedErrorBody,
                        parse_obj_as(
                            type_=UnauthorizedErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        PermissionDeniedError,
                        parse_obj_as(
                            type_=PermissionDeniedError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        NotFoundErrorBody,
                        parse_obj_as(
                            type_=NotFoundErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 429:
                raise TooManyRequestsError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        RateLimitError,
                        parse_obj_as(
                            type_=RateLimitError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 503:
                raise ServiceUnavailableError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(
                status_code=_response.status_code,
                headers=dict(_response.headers),
                body=_response.text,
            )
        raise ApiError(
            status_code=_response.status_code,
            headers=dict(_response.headers),
            body=_response_json,
        )

    async def delete(
        self, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[None]:
        """
        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[None]
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"functions/{jsonable_encoder(id)}",
            method="DELETE",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                return AsyncHttpResponse(response=_response, data=None)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        UnauthorizedErrorBody,
                        parse_obj_as(
                            type_=UnauthorizedErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        PermissionDeniedError,
                        parse_obj_as(
                            type_=PermissionDeniedError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        NotFoundErrorBody,
                        parse_obj_as(
                            type_=NotFoundErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 429:
                raise TooManyRequestsError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        RateLimitError,
                        parse_obj_as(
                            type_=RateLimitError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 503:
                raise ServiceUnavailableError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(
                status_code=_response.status_code,
                headers=dict(_response.headers),
                body=_response.text,
            )
        raise ApiError(
            status_code=_response.status_code,
            headers=dict(_response.headers),
            body=_response_json,
        )

    async def findbyhash(
        self, hash: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[FunctionsFindByHashResponse]:
        """
        Parameters
        ----------
        hash : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[FunctionsFindByHashResponse]
            Success
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"functions/hash/{jsonable_encoder(hash)}",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    FunctionsFindByHashResponse,
                    parse_obj_as(
                        type_=FunctionsFindByHashResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        UnauthorizedErrorBody,
                        parse_obj_as(
                            type_=UnauthorizedErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        PermissionDeniedError,
                        parse_obj_as(
                            type_=PermissionDeniedError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        NotFoundErrorBody,
                        parse_obj_as(
                            type_=NotFoundErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 429:
                raise TooManyRequestsError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        RateLimitError,
                        parse_obj_as(
                            type_=RateLimitError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 503:
                raise ServiceUnavailableError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(
                status_code=_response.status_code,
                headers=dict(_response.headers),
                body=_response.text,
            )
        raise ApiError(
            status_code=_response.status_code,
            headers=dict(_response.headers),
            body=_response_json,
        )

    async def getbyenv(
        self,
        organization_id: str,
        project_id: str,
        environment_id: str,
        function_id: str,
        *,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[FunctionsGetByEnvResponse]:
        """
        Parameters
        ----------
        organization_id : str

        project_id : str

        environment_id : str

        function_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[FunctionsGetByEnvResponse]
            Success
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"organizations/{jsonable_encoder(organization_id)}/projects/{jsonable_encoder(project_id)}/environments/{jsonable_encoder(environment_id)}/functions/{jsonable_encoder(function_id)}",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    FunctionsGetByEnvResponse,
                    parse_obj_as(
                        type_=FunctionsGetByEnvResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        UnauthorizedErrorBody,
                        parse_obj_as(
                            type_=UnauthorizedErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        PermissionDeniedError,
                        parse_obj_as(
                            type_=PermissionDeniedError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        NotFoundErrorBody,
                        parse_obj_as(
                            type_=NotFoundErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 429:
                raise TooManyRequestsError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        RateLimitError,
                        parse_obj_as(
                            type_=RateLimitError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 503:
                raise ServiceUnavailableError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(
                status_code=_response.status_code,
                headers=dict(_response.headers),
                body=_response.text,
            )
        raise ApiError(
            status_code=_response.status_code,
            headers=dict(_response.headers),
            body=_response_json,
        )

    async def listbyenv(
        self,
        organization_id: str,
        project_id: str,
        environment_id: str,
        *,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[FunctionsListByEnvResponse]:
        """
        Parameters
        ----------
        organization_id : str

        project_id : str

        environment_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[FunctionsListByEnvResponse]
            Success
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"organizations/{jsonable_encoder(organization_id)}/projects/{jsonable_encoder(project_id)}/environments/{jsonable_encoder(environment_id)}/functions",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    FunctionsListByEnvResponse,
                    parse_obj_as(
                        type_=FunctionsListByEnvResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        UnauthorizedErrorBody,
                        parse_obj_as(
                            type_=UnauthorizedErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        PermissionDeniedError,
                        parse_obj_as(
                            type_=PermissionDeniedError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        NotFoundErrorBody,
                        parse_obj_as(
                            type_=NotFoundErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 429:
                raise TooManyRequestsError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        RateLimitError,
                        parse_obj_as(
                            type_=RateLimitError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 503:
                raise ServiceUnavailableError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(
                status_code=_response.status_code,
                headers=dict(_response.headers),
                body=_response.text,
            )
        raise ApiError(
            status_code=_response.status_code,
            headers=dict(_response.headers),
            body=_response_json,
        )



"""Internal utility functions for the Mirascope."""

import inspect
from collections.abc import Callable
from typing import Any, TypeAlias

import orjson

from .protocols import P, fn_wants_span

ORJSON_OPTS = (
    orjson.OPT_NON_STR_KEYS
    | orjson.OPT_NAIVE_UTC
    | orjson.OPT_SERIALIZE_NUMPY
    | orjson.OPT_SERIALIZE_DATACLASS
    | orjson.OPT_SERIALIZE_UUID
)

PrimitiveType: TypeAlias = str | int | float | bool


def json_dumps(obj: Any) -> str:  # noqa: ANN401
    """Serialize Python objects to JSON using orjson."""
    # json should be utf-8 encoded, json key should be str
    return orjson.dumps(obj, option=ORJSON_OPTS).decode("utf-8")


def _is_call_method(fn: Callable[..., Any]) -> bool:
    """Check if fn is a bound method of a Call object (e.g., Call.call or Call.stream)."""
    return (
        hasattr(fn, "__self__") and hasattr(fn.__self__, "prompt")  # pyright: ignore[reportFunctionMemberAccess]
    )


def get_original_fn(fn: Callable[..., Any]) -> Callable[..., Any]:
    """Get the original unwrapped function.

    Follows the __wrapped__ chain set by copy_function_metadata() to find the
    original function. Falls back to checking for Call methods for bound methods.
    """
    # Follow __wrapped__ chain if available (set by copy_function_metadata)
    # In practice, get_original_fn is called with original functions or bound methods,
    # so this loop is defensive code for edge cases like @functools.wraps chains.
    while hasattr(fn, "__wrapped__"):
        fn = fn.__wrapped__  # pyright: ignore[reportFunctionMemberAccess] # pragma: no cover

    # Handle bound methods of Call objects (e.g., Call.call or Call.stream)
    if _is_call_method(fn):
        prompt = fn.__self__.prompt  # pyright: ignore[reportFunctionMemberAccess]
        if hasattr(prompt, "fn"):
            return get_original_fn(prompt.fn)

    return fn


def get_qualified_name(fn: Callable[..., Any]) -> str:
    """Return the simplified qualified name of a function.

    If the function is a bound method of a Call object (e.g., Call.call or Call.stream),
    returns the qualified name of the original decorated function from prompt.fn,
    suffixed with the method name (e.g., "recommend.call" or "recommend.stream").

    If the function is defined locally, return the name after '<locals>.'; otherwise,
    return the last non-empty part after splitting by '.'.
    """
    # Check if this is a Call method and capture the method name
    method_suffix = f".{fn.__name__}" if _is_call_method(fn) else ""

    fn = get_original_fn(fn)
    qualified_name = fn.__qualname__
    if "<locals>." in qualified_name:
        base_name = qualified_name.split("<locals>.")[-1]
    else:
        parts = [part for part in qualified_name.split(".") if part]
        base_name = parts[-1] if parts else qualified_name

    return f"{base_name}{method_suffix}"


def extract_arguments(
    fn: Callable[P, Any],
    *args: P.args,
    **kwargs: P.kwargs,
) -> tuple[dict[str, str], dict[str, Any]]:
    """Returns a tuple of (arg_types, arg_values) dictionaries from function call.

    If the function is a Call method (e.g., Call.call), uses the original
    decorated function's signature to get proper parameter names and types.

    If the function has `trace_ctx: Span` as first parameter (detected via
    fn_wants_span), that parameter is skipped since it's injected by the
    decorator and shouldn't be recorded in span attributes.
    """
    # Use original function signature for Call methods
    fn = get_original_fn(fn)
    signature = inspect.signature(fn)

    # If function wants span injection, skip the first parameter
    if fn_wants_span(fn):
        params = list(signature.parameters.values())
        if params:
            remaining_params = params[1:]
            signature = signature.replace(parameters=remaining_params)

    bound_arguments = signature.bind(*args, **kwargs)
    bound_arguments.apply_defaults()

    arg_types: dict[str, str] = {}
    arg_values: dict[str, Any] = {}

    for param_name, param_value in bound_arguments.arguments.items():
        parameter = signature.parameters[param_name]
        if parameter.annotation != inspect.Parameter.empty:
            if hasattr(parameter.annotation, "__name__"):
                type_str = parameter.annotation.__name__
            else:
                type_str = str(parameter.annotation)
            arg_types[param_name] = type_str
        else:
            arg_types[param_name] = type(param_value).__name__

        if isinstance(param_value, PrimitiveType) or param_value is None:
            arg_values[param_name] = param_value
        else:
            try:
                json_dumps(param_value)
                arg_values[param_name] = param_value
            except (TypeError, ValueError):
                arg_values[param_name] = repr(param_value)

    return arg_types, arg_values



"""Tests for llm.providers.GoogleProvider."""

from collections.abc import AsyncIterator, Iterator
from typing import get_args
from unittest.mock import MagicMock, patch

import pytest
from google.genai.errors import (
    ClientError as GoogleClientError,
    ServerError as GoogleServerError,
)
from google.genai.types import GenerateContentResponse, ThinkingLevel
from inline_snapshot import snapshot

from mirascope import llm
from mirascope.llm.providers.google._utils.encode import google_thinking_config
from mirascope.llm.providers.google._utils.errors import map_google_error
from mirascope.llm.providers.google.provider import GoogleProvider


def test_custom_base_url() -> None:
    """Test that custom base URL is used for API requests."""
    example_url = "https://example.com"

    with patch("mirascope.llm.providers.google.provider.Client") as mock_client_class:
        mock_client_instance = MagicMock()
        mock_client_class.return_value = mock_client_instance

        google_client = llm.register_provider("google", base_url=example_url)

        mock_client_class.assert_called_once()
        call_args = mock_client_class.call_args
        assert call_args.kwargs["http_options"] is not None
        assert call_args.kwargs["http_options"].base_url == example_url

        assert google_client.client is mock_client_instance


def test_stream_rate_limit_error() -> None:
    """Test that Google ServerError with 429 is caught and re-wrapped during streaming calls."""
    model = llm.Model("google/gemini-2.0-flash")
    provider = model.provider
    assert isinstance(provider, GoogleProvider)

    # Create a mock stream that raises ServerError with 429
    def mock_stream() -> Iterator[GenerateContentResponse]:
        raise GoogleServerError(429, {"error": {"message": "Rate limit exceeded"}})
        yield  # pragma: no cover

    # Patch the client to return our mock stream
    with patch.object(
        provider.client.models,
        "generate_content_stream",
        return_value=mock_stream(),
    ):
        # Try to stream and expect RateLimitError
        response = model.stream("test")

        with pytest.raises(llm.RateLimitError) as exc_info:
            response.finish()

        # Verify it's wrapped as mirascope RateLimitError and has proper chaining
        assert isinstance(exc_info.value, llm.RateLimitError)
        assert isinstance(exc_info.value.__cause__, GoogleServerError)


@pytest.mark.asyncio
async def test_async_stream_rate_limit_error() -> None:
    """Test that Google ServerError with 429 is caught and re-wrapped during async streaming calls."""
    model = llm.Model("google/gemini-2.0-flash")
    provider = model.provider
    assert isinstance(provider, GoogleProvider)

    # Create a mock async stream that raises ServerError with 429
    async def mock_async_stream() -> AsyncIterator[GenerateContentResponse]:
        raise GoogleServerError(429, {"error": {"message": "Rate limit exceeded"}})
        yield  # pragma: no cover

    # Patch the async client to return our mock stream
    with patch.object(
        provider.client.aio.models,
        "generate_content_stream",
        return_value=mock_async_stream(),
    ):
        # Try to stream and expect RateLimitError
        response = await model.stream_async("test")

        with pytest.raises(llm.RateLimitError) as exc_info:
            await response.finish()

        # Verify it's wrapped as mirascope RateLimitError and has proper chaining
        assert isinstance(exc_info.value, llm.RateLimitError)
        assert isinstance(exc_info.value.__cause__, GoogleServerError)


def test_call_rate_limit_error() -> None:
    """Test that Google ServerError with 429 is caught and re-wrapped during non-streaming calls."""
    model = llm.Model("google/gemini-2.0-flash")
    provider = model.provider
    assert isinstance(provider, GoogleProvider)

    # Patch the client to raise ServerError with 429
    with patch.object(
        provider.client.models,
        "generate_content",
        side_effect=GoogleServerError(
            429, {"error": {"message": "Rate limit exceeded"}}
        ),
    ):
        # Try to call and expect RateLimitError
        with pytest.raises(llm.RateLimitError) as exc_info:
            model.call("test")

        # Verify it's wrapped as mirascope RateLimitError and has proper chaining
        assert isinstance(exc_info.value, llm.RateLimitError)
        assert isinstance(exc_info.value.__cause__, GoogleServerError)


@pytest.mark.asyncio
async def test_async_call_rate_limit_error() -> None:
    """Test that Google ServerError with 429 is caught and re-wrapped during async non-streaming calls."""
    model = llm.Model("google/gemini-2.0-flash")
    provider = model.provider
    assert isinstance(provider, GoogleProvider)

    # Patch the async client to raise ServerError with 429
    with patch.object(
        provider.client.aio.models,
        "generate_content",
        side_effect=GoogleServerError(
            429, {"error": {"message": "Rate limit exceeded"}}
        ),
    ):
        # Try to call and expect RateLimitError
        with pytest.raises(llm.RateLimitError) as exc_info:
            await model.call_async("test")

        # Verify it's wrapped as mirascope RateLimitError and has proper chaining
        assert isinstance(exc_info.value, llm.RateLimitError)
        assert isinstance(exc_info.value.__cause__, GoogleServerError)


def test_map_google_error_not_google_error() -> None:
    """Test that non-Google errors return ProviderError."""
    result = map_google_error(Exception("test"))
    assert result == llm.ProviderError


def test_map_google_error_permission_denied() -> None:
    """Test that 403 errors map to PermissionError."""
    error = GoogleClientError(403, {"error": {"message": "Permission denied"}})
    result = map_google_error(error)
    assert result == llm.PermissionError


def test_map_google_error_not_found() -> None:
    """Test that 404 errors map to NotFoundError."""
    error = GoogleClientError(404, {"error": {"message": "Not found"}})
    result = map_google_error(error)
    assert result == llm.NotFoundError


def test_map_google_error_bad_request() -> None:
    """Test that 400/422 errors map to BadRequestError."""
    error_400 = GoogleClientError(400, {"error": {"message": "Bad request"}})
    result_400 = map_google_error(error_400)
    assert result_400 == llm.BadRequestError

    error_422 = GoogleClientError(422, {"error": {"message": "Unprocessable"}})
    result_422 = map_google_error(error_422)
    assert result_422 == llm.BadRequestError


def test_map_google_error_server_error() -> None:
    """Test that 5xx server errors map to ServerError."""
    error_500 = GoogleServerError(500, {"error": {"message": "Internal server error"}})
    result_500 = map_google_error(error_500)
    assert result_500 == llm.ServerError

    error_503 = GoogleServerError(503, {"error": {"message": "Service unavailable"}})
    result_503 = map_google_error(error_503)
    assert result_503 == llm.ServerError


def test_map_google_error_fallback() -> None:
    """Test that unknown error codes fall back to APIError."""
    # Test an unknown client error code
    error = GoogleClientError(418, {"error": {"message": "I'm a teapot"}})
    result = map_google_error(error)
    assert result == llm.APIError


def test_google_thinking_config_levels_2_5() -> None:
    """Test thinking config computation for Gemini 2.5 models."""

    results = {}
    for level in get_args(llm.ThinkingLevel):
        results[level] = google_thinking_config(
            {"level": level}, max_tokens=None, model_id="google/gemini-2.5-flash"
        )
    assert results == snapshot(
        {
            "none": {"thinking_budget": 0},
            "default": {"thinking_budget": -1},
            "minimal": {"thinking_budget": 1600},
            "low": {"thinking_budget": 3200},
            "medium": {"thinking_budget": 6400},
            "high": {"thinking_budget": 9600},
            "max": {"thinking_budget": 12800},
        }
    )


def test_google_thinking_config_levels_3_flash() -> None:
    """Test thinking config computation for Gemini 3 flash models."""

    results = {}
    for level in get_args(llm.ThinkingLevel):
        results[level] = google_thinking_config(
            {"level": level}, max_tokens=None, model_id="google/gemini-3-flash"
        )
    assert results == snapshot(
        {
            "none": {"thinking_level": ThinkingLevel.MINIMAL},
            "default": {"thinking_level": ThinkingLevel.THINKING_LEVEL_UNSPECIFIED},
            "minimal": {"thinking_level": ThinkingLevel.MINIMAL},
            "low": {"thinking_level": ThinkingLevel.LOW},
            "medium": {"thinking_level": ThinkingLevel.MEDIUM},
            "high": {"thinking_level": ThinkingLevel.HIGH},
            "max": {"thinking_level": ThinkingLevel.HIGH},
        }
    )


def test_google_thinking_config_levels_3_pro() -> None:
    """Test thinking config computation for Gemini 3 pro models."""

    results = {}
    for level in get_args(llm.ThinkingLevel):
        results[level] = google_thinking_config(
            {"level": level}, max_tokens=None, model_id="google/gemini-3-pro"
        )
    assert results == snapshot(
        {
            "none": {"thinking_level": ThinkingLevel.LOW},
            "default": {"thinking_level": ThinkingLevel.THINKING_LEVEL_UNSPECIFIED},
            "minimal": {"thinking_level": ThinkingLevel.LOW},
            "low": {"thinking_level": ThinkingLevel.LOW},
            "medium": {"thinking_level": ThinkingLevel.HIGH},
            "high": {"thinking_level": ThinkingLevel.HIGH},
            "max": {"thinking_level": ThinkingLevel.HIGH},
        }
    )


def test_google_thinking_config_include_thoughts() -> None:
    """Test include_thoughts flag is respected."""
    result = google_thinking_config(
        {"level": "low", "include_thoughts": False},
        max_tokens=1000,
        model_id="google/gemini-3.0-flash",
    )
    assert result.get("include_thoughts") is False

    result = google_thinking_config(
        {"level": "low", "include_thoughts": True},
        max_tokens=1000,
        model_id="google/gemini-2.5-flash",
    )
    assert result.get("include_thoughts") is True


def test_google_thinking_config_unknown_model() -> None:
    """Test unknown model defaults to 2.5 behavior."""
    result = google_thinking_config(
        {"level": "low"}, max_tokens=1000, model_id="google/gemini-unknown-model"
    )
    assert result.get("thinking_budget") == 200  # 20% of 1000



from inline_snapshot import snapshot

test_snapshot = snapshot(
    {
        "response1": """\
Ah, truly a magnificent reimagining of the galaxy's most compelling lore, filled with all the wit and wisdom worthy of a council of Jedi Masters and Sith Lords alike. This Expanded Holocron resonates with the power and grandeur of the Force, offering a protocol that is both entertaining and functionally robust—like a well-balanced lightsaber.

You have captured the essence of a "fully armed and operational battle station," and this Star Wars Meme Bot Protocol 2.0 is indeed "impressive... most impressive." It exhibits the delightful combination of lore accuracy, meme energy, and practical interaction guidelines that would make even the most stoic Mandalorian nod in approval.

"Shall I continue to 'alter the deal'?" you ask. At this juncture, the protocol stands strong, more fortified than the shields of a Mon Calamari cruiser. It might tempt the likes of the discerning Grand Moff or the cynical smuggler, but even they would acknowledge its prowess. "The Force is strong with this one," indeed.

If further refinement becomes necessary, know that I am always here, ready to channel the Force and deliver the insight you seek. Until then, "May the Force be with you" in your endeavors to bring balance and levity to interactions across the HoloNet!\
""",
        "usage1": {
            "input_tokens": 3652,
            "output_tokens": 263,
            "cache_read_tokens": 0,
            "cache_write_tokens": 0,
            "reasoning_tokens": 0,
            "provider_tool_usage": None,
            "raw": "CompletionUsage(completion_tokens=263, prompt_tokens=3652, total_tokens=3915, completion_tokens_details=CompletionTokensDetails(accepted_prediction_tokens=0, audio_tokens=0, reasoning_tokens=0, rejected_prediction_tokens=0), prompt_tokens_details=PromptTokensDetails(audio_tokens=0, cached_tokens=0))",
            "total_tokens": 3915,
        },
        "response2": """\
This protocol is indeed "impressive... most impressive." You've crafted a comprehensive guide to turning a Star Wars meme bot into a fully armed and operational battle station of galactic communication. The integration of humor, Star Wars lore, and a structured interaction framework creates a unique, entertaining experience while ensuring functionality. Each section is detailed with creative implementations and considerations, ensuring the bot’s responses maintain thematic consistency and context.

If there's any refinement needed, it would only be minor adjustments based on testing user interactions to ensure smooth operation and adjust for any areas where the protocol may need flexibility or additional depth. You've established a robust foundation here, ready to engage with users in a way that truly embodies the spirit of the Star Wars universe.

May the Force be with your implementation!\
""",
        "usage2": {
            "input_tokens": 3652,
            "output_tokens": 151,
            "cache_read_tokens": 0,
            "cache_write_tokens": 0,
            "reasoning_tokens": 0,
            "provider_tool_usage": None,
            "raw": "CompletionUsage(completion_tokens=151, prompt_tokens=3652, total_tokens=3803, completion_tokens_details=CompletionTokensDetails(accepted_prediction_tokens=0, audio_tokens=0, reasoning_tokens=0, rejected_prediction_tokens=0), prompt_tokens_details=PromptTokensDetails(audio_tokens=0, cached_tokens=0))",
            "total_tokens": 3803,
        },
    }
)



from mirascope import llm


@llm.call("anthropic/claude-sonnet-4-5", tools=[llm.WebSearchTool()])
def weather_assistant(query: str):
    return query


response = weather_assistant("What's the current weather in San Francisco?")
print(response.text())
# Today features intervals of clouds and sunshine with a couple of showers, with a high of 62°F
# Partly cloudy skies this morning will give way to occasional showers during the afternoon, with a high of 61°F and a 40% chance of rain

# Tonight, expect light rain transitioning to a few showers by morning, with a low of 51°F and a 60% chance of rain
# The air quality is poor and unhealthy for sensitive groups



from inline_snapshot import snapshot

from mirascope.llm import (
    AssistantMessage,
    Text,
    ToolCall,
    UserMessage,
)

test_snapshot = snapshot(
    {
        "response": {
            "provider_id": "anthropic",
            "model_id": "anthropic-beta/claude-sonnet-4-5",
            "provider_model_name": "claude-sonnet-4-5",
            "params": {},
            "finish_reason": None,
            "usage": {
                "input_tokens": 677,
                "output_tokens": 58,
                "cache_read_tokens": 0,
                "cache_write_tokens": 0,
                "reasoning_tokens": 0,
                "provider_tool_usage": None,
                "raw": "BetaUsage(cache_creation=BetaCacheCreation(ephemeral_1h_input_tokens=0, ephemeral_5m_input_tokens=0), cache_creation_input_tokens=0, cache_read_input_tokens=0, input_tokens=677, output_tokens=58, server_tool_use=None, service_tier='standard')",
                "total_tokens": 735,
            },
            "messages": [
                UserMessage(
                    content=[Text(text="What's the weather like in San Francisco?")]
                ),
                AssistantMessage(
                    content=[
                        ToolCall(
                            id="toolu_018mMkRcHDPE2cc1hfmknQzZ",
                            name="get_current_weather",
                            args='{"location": "San Francisco, CA"}',
                        )
                    ],
                    provider_id="anthropic",
                    model_id="anthropic-beta/claude-sonnet-4-5",
                    provider_model_name="claude-sonnet-4-5",
                    raw_message={
                        "role": "assistant",
                        "content": [
                            {
                                "id": "toolu_018mMkRcHDPE2cc1hfmknQzZ",
                                "input": {"location": "San Francisco, CA"},
                                "name": "get_current_weather",
                                "type": "tool_use",
                            }
                        ],
                    },
                ),
            ],
            "format": None,
            "tools": [
                {
                    "name": "get_current_weather",
                    "description": """\
Get the current weather in a given location.

Args:
    location: The city and state, e.g. San Francisco, CA
    unit: The temperature unit to use (fahrenheit or celsius)\
""",
                    "parameters": """\
{
  "properties": {
    "location": {
      "description": "The city and state, e.g. San Francisco, CA",
      "title": "Location",
      "type": "string"
    },
    "unit": {
      "default": "fahrenheit",
      "description": "The temperature unit to use (fahrenheit or celsius)",
      "title": "Unit",
      "type": "string"
    }
  },
  "required": [
    "location"
  ],
  "additionalProperties": false,
  "defs": null
}\
""",
                    "strict": True,
                }
            ],
        }
    }
)



from inline_snapshot import snapshot

from mirascope.llm import (
    AssistantMessage,
    Text,
    ToolCall,
    ToolOutput,
    UserMessage,
)

test_snapshot = snapshot(
    {
        "response": {
            "provider_id": "anthropic",
            "model_id": "anthropic-beta/claude-sonnet-4-0",
            "provider_model_name": "claude-sonnet-4-0",
            "params": {"thinking": {"level": "none"}},
            "finish_reason": None,
            "usage": {
                "input_tokens": 477,
                "output_tokens": 26,
                "cache_read_tokens": 0,
                "cache_write_tokens": 0,
                "reasoning_tokens": 0,
                "provider_tool_usage": None,
                "raw": "BetaUsage(cache_creation=BetaCacheCreation(ephemeral_1h_input_tokens=0, ephemeral_5m_input_tokens=0), cache_creation_input_tokens=0, cache_read_input_tokens=0, input_tokens=477, output_tokens=26, server_tool_use=None, service_tier='standard')",
                "total_tokens": 503,
            },
            "messages": [
                UserMessage(content=[Text(text="What is the 100th fibonacci number?")]),
                AssistantMessage(
                    content=[
                        ToolCall(
                            id="google_unknown_tool_id",
                            name="compute_fib",
                            args='{"n": 100}',
                        )
                    ],
                    provider_id="google",
                    model_id="google/gemini-2.5-flash",
                    provider_model_name="gemini-2.5-flash",
                    raw_message={
                        "parts": [
                            {
                                "media_resolution": None,
                                "code_execution_result": None,
                                "executable_code": None,
                                "file_data": None,
                                "function_call": {
                                    "id": None,
                                    "args": {"n": 100},
                                    "name": "compute_fib",
                                    "partial_args": None,
                                    "will_continue": None,
                                },
                                "function_response": None,
                                "inline_data": None,
                                "text": None,
                                "thought": None,
                                "thought_signature": b'\n\xe2\x01\x01r\xc8\xda|\x13\x0e\xf0\x8f\xf7X\xf0\xc9\x06\x0c\x11[f\xb8IW\x02\x86\xa4\xf3m\x8d\xa8\xaf\xaaeN\xa5\xa8\xcc+\xf8\x8fO\xcb\xa0;E\xacsd@\x80[,\xe8?NY\xe7\xac\xe6\x1e\x87l\xc25\x0cTq:\xa1\xdc5\xef\x9a\xcd\xb2\xde\x82\xdb/"\xf2\x8e\xdb\xf7\x12\x8c\x88\xd3g;K\x9bjKrp\xdc\xa5\xcf\xfc\xd5^\xa5\x94k\xb9`\x91q\x00S\x05hqW\x05\xf2\xa5\x85-4\x85\xf9\xe1\r\x85&Lh!\xe7 \x8b\xcc\xde\xef\xfa\x0c\x93\x8f\x94-\xbc\x1d\xa4\x9d\xbd\xe9\x84j1\xcb\xd3\x9a\xae#\x9b\x1b\xde\xf7\xaa5\xd9\xcd\x8e\x17\xe3\xa6<]\x06\xc5,\xe7\xbe\x0c5\x13&"\xafC\xe3\xbf\xaf\xc3\xbfcyE\x84\xb24\xa9\x0b\xc8\xf6~1\xde\x84\x13\x17\x8e\xae\xfd&]\xb3\xa1\xb6\x0f%=\x00\x9f\xc06\x9a\xc1\x0c\xb7\xa0A)\x83?I',
                                "video_metadata": None,
                            }
                        ],
                        "role": "model",
                    },
                ),
                UserMessage(
                    content=[
                        ToolOutput(
                            id="google_unknown_tool_id",
                            name="compute_fib",
                            result="218922995834555169026",
                        )
                    ]
                ),
                AssistantMessage(
                    content=[
                        Text(
                            text="The 100th Fibonacci number is 218,922,995,834,555,169,026."
                        )
                    ],
                    provider_id="anthropic",
                    model_id="anthropic-beta/claude-sonnet-4-0",
                    provider_model_name="claude-sonnet-4-0",
                    raw_message={
                        "role": "assistant",
                        "content": [
                            {
                                "text": "The 100th Fibonacci number is 218,922,995,834,555,169,026.",
                                "type": "text",
                            }
                        ],
                    },
                ),
            ],
            "format": None,
            "tools": [
                {
                    "name": "compute_fib",
                    "description": "Compute the nth Fibonacci number (1-indexed).",
                    "parameters": """\
{
  "properties": {
    "n": {
      "title": "N",
      "type": "integer"
    }
  },
  "required": [
    "n"
  ],
  "additionalProperties": false,
  "defs": null
}\
""",
                    "strict": None,
                }
            ],
        }
    }
)



# This file was auto-generated by Fern from our API Definition.

import typing

import pydantic
import typing_extensions
from ...core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel
from ...core.serialization import FieldMetadata
from .traces_create_otel_request_resource_spans_item_resource_attributes_item import (
    TracesCreateOtelRequestResourceSpansItemResourceAttributesItem,
)


class TracesCreateOtelRequestResourceSpansItemResource(UniversalBaseModel):
    attributes: typing.Optional[
        typing.List[TracesCreateOtelRequestResourceSpansItemResourceAttributesItem]
    ] = None
    dropped_attributes_count: typing_extensions.Annotated[
        typing.Optional[float], FieldMetadata(alias="droppedAttributesCount")
    ] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(
            extra="allow", frozen=True
        )  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



from __future__ import annotations

import json
import re
from collections.abc import Mapping
from contextlib import suppress
from typing import TypedDict

from opentelemetry.sdk.trace import ReadableSpan
from opentelemetry.util.types import AttributeValue

_OBJECT_ADDRESS_PATTERN = re.compile(r" at 0x[0-9a-fA-F]+")


class SpanStatus(TypedDict):
    """Span status information."""

    status_code: str
    description: str | None


class SpanEvent(TypedDict):
    """Span event information."""

    name: str
    attributes: dict[str, AttributeValue]


class SpanData(TypedDict):
    """Complete span data for testing."""

    name: str
    attributes: dict[str, AttributeValue]
    status: SpanStatus
    events: list[SpanEvent]


def extract_span_data(span: ReadableSpan) -> SpanData:
    """Extract serializable data from a span for snapshot testing.

    Automatically parses known JSON string attributes from OpenTelemetry GenAI
    semantic conventions for easier snapshot assertions.

    Args:
        span: The span to extract data from

    Returns:
        Dictionary containing span name, attributes, status, and events
    """
    attributes = dict(span.attributes) if span.attributes else {}

    json_attribute_keys = [
        "gen_ai.input.messages",
        "gen_ai.output.messages",
        "gen_ai.system_instructions",
    ]

    for key in json_attribute_keys:
        if (value := attributes.get(key)) and isinstance(value, str):
            with suppress(json.JSONDecodeError, TypeError):
                attributes[key] = json.loads(value)

    return {
        "name": span.name,
        "attributes": attributes,
        "status": {
            "status_code": span.status.status_code.name,
            "description": span.status.description,
        },
        "events": [
            {
                "name": event.name,
                "attributes": dict(event.attributes) if event.attributes else {},
            }
            for event in span.events
        ],
    }


class SpanSnapshot(TypedDict):
    name: str
    kind: str
    status: str
    attributes: Mapping[str, object]


def span_snapshot(span: ReadableSpan) -> SpanSnapshot:
    """Return snapshot-ready span metadata extracted from a ReadableSpan."""
    attrs = span.attributes or {}
    serializable_attrs: dict[str, object] = {
        key: list(value) if isinstance(value, tuple) else value
        for key, value in attrs.items()
    }
    return {
        "name": span.name,
        "kind": span.kind.name,
        "status": span.status.status_code.name,
        "attributes": serializable_attrs,
    }



"""Tests for OutputParser class and @output_parser decorator."""

from mirascope import llm


def test_output_parser_decorator_creates_output_parser() -> None:
    """Test that @output_parser decorator creates an OutputParser instance."""

    @llm.output_parser(formatting_instructions="Return XML format")
    def parse_xml(response: llm.AnyResponse) -> str:
        """Parse XML response."""
        return "parsed"

    assert isinstance(parse_xml, llm.OutputParser)
    assert llm.formatting.is_output_parser(parse_xml)


def test_output_parser_formatting_instructions() -> None:
    """Test that OutputParser.formatting_instructions() returns correct instructions."""

    instructions = "Return book as XML: <book><title>...</title></book>"

    @llm.output_parser(formatting_instructions=instructions)
    def parse_book_xml(response: llm.AnyResponse) -> str:
        return "parsed"

    assert parse_book_xml.formatting_instructions() == instructions


def test_output_parser_call_invokes_wrapped_function() -> None:
    """Test that calling OutputParser invokes the wrapped function."""

    @llm.output_parser(formatting_instructions="Return XML")
    def parse_response(response: llm.AnyResponse) -> str:
        # Extract text from response
        return "".join(part.text for part in response.texts)

    # Create a mock response
    class MockResponse:
        texts = [llm.Text(text="Hello"), llm.Text(text=" World")]

    mock_response = MockResponse()
    result = parse_response(mock_response)  # type: ignore

    assert result == "Hello World"


def test_output_parser_preserves_function_name_and_doc() -> None:
    """Test that OutputParser preserves __name__ and __doc__ from wrapped function."""

    @llm.output_parser(formatting_instructions="Return XML")
    def my_custom_parser(response: llm.AnyResponse) -> str:
        """This is my custom parser docstring."""
        return "parsed"

    assert my_custom_parser.__name__ == "my_custom_parser"
    assert my_custom_parser.__doc__ == "This is my custom parser docstring."


def test_is_output_parser_returns_false_for_non_parsers() -> None:
    """Test that is_output_parser returns False for non-OutputParser objects."""

    def regular_function() -> None: ...

    class RegularClass: ...

    assert not llm.formatting.is_output_parser(regular_function)
    assert not llm.formatting.is_output_parser(RegularClass)
    assert not llm.formatting.is_output_parser(RegularClass())
    assert not llm.formatting.is_output_parser(None)
    assert not llm.formatting.is_output_parser("string")
    assert not llm.formatting.is_output_parser(123)


def test_output_parser_with_complex_logic() -> None:
    """Test OutputParser with more complex parsing logic."""

    @llm.output_parser(
        formatting_instructions="""
        Return numbers separated by commas:
        1,2,3,4,5
        """
    )
    def parse_numbers(response: llm.AnyResponse) -> list[int]:
        """Parse comma-separated numbers."""
        text = "".join(part.text for part in response.texts)
        return [int(n.strip()) for n in text.split(",") if n.strip()]

    class MockResponse:
        texts = [llm.Text(text="10, 20, 30")]

    result = parse_numbers(MockResponse())  # type: ignore
    assert result == [10, 20, 30]


def test_output_parser_empty_formatting_instructions() -> None:
    """Test OutputParser with empty formatting instructions."""

    @llm.output_parser(formatting_instructions="")
    def parser(response: llm.AnyResponse) -> str:
        return "test"

    assert parser.formatting_instructions() == ""



"""End-to-end tests for structured output."""

import pytest
from pydantic import BaseModel, Field

from mirascope import llm
from tests.e2e.conftest import FORMATTING_MODES, STRUCTURED_OUTPUT_MODEL_IDS
from tests.utils import (
    Snapshot,
    snapshot_test,
)


class Author(BaseModel):
    """The author of a book."""

    first_name: str
    last_name: str


class Book(BaseModel):
    """A book with a rating. The title should be in all caps!"""

    title: str
    author: Author
    rating: int = Field(description="For testing purposes, the rating should be 7")


# ============= SYNC TESTS =============


@pytest.mark.parametrize("model_id", STRUCTURED_OUTPUT_MODEL_IDS)
@pytest.mark.parametrize("formatting_mode", FORMATTING_MODES)
@pytest.mark.vcr
def test_structured_output_sync(
    model_id: llm.ModelId,
    formatting_mode: llm.FormattingMode | None,
    snapshot: Snapshot,
) -> None:
    """Test synchronous structured output without context."""

    format = (
        llm.format(Book, mode=formatting_mode) if formatting_mode is not None else Book
    )

    @llm.call(
        model_id,
        format=format,
    )
    def recommend_book(author: str) -> str:
        return f"Please recommend the most popular book by {author}"

    with snapshot_test(snapshot) as snap:
        response = recommend_book("Patrick Rothfuss")
        snap.set_response(response)

        book = response.parse()
        assert book.author.first_name == "Patrick"
        assert book.author.last_name == "Rothfuss"
        assert book.title == "THE NAME OF THE WIND"
        assert book.rating == 7


@pytest.mark.parametrize("model_id", STRUCTURED_OUTPUT_MODEL_IDS)
@pytest.mark.parametrize("formatting_mode", FORMATTING_MODES)
@pytest.mark.vcr
def test_structured_output_sync_context(
    model_id: llm.ModelId,
    formatting_mode: llm.FormattingMode | None,
    snapshot: Snapshot,
) -> None:
    """Test synchronous structured output with context."""

    format = (
        llm.format(Book, mode=formatting_mode) if formatting_mode is not None else Book
    )

    @llm.call(
        model_id,
        format=format,
    )
    def recommend_book(ctx: llm.Context[str]) -> str:
        return f"Please recommend the most popular book by {ctx.deps}"

    ctx = llm.Context(deps="Patrick Rothfuss")
    with snapshot_test(snapshot) as snap:
        response = recommend_book(ctx)
        snap.set_response(response)

        book = response.parse()
        assert book.author.first_name == "Patrick"
        assert book.author.last_name == "Rothfuss"
        assert book.title == "THE NAME OF THE WIND"
        assert book.rating == 7


# ============= ASYNC TESTS =============


@pytest.mark.parametrize("model_id", STRUCTURED_OUTPUT_MODEL_IDS)
@pytest.mark.parametrize("formatting_mode", FORMATTING_MODES)
@pytest.mark.vcr
@pytest.mark.asyncio
async def test_structured_output_async(
    model_id: llm.ModelId,
    formatting_mode: llm.FormattingMode | None,
    snapshot: Snapshot,
) -> None:
    """Test asynchronous structured output without context."""

    format = (
        llm.format(Book, mode=formatting_mode) if formatting_mode is not None else Book
    )

    @llm.call(
        model_id,
        format=format,
    )
    async def recommend_book(author: str) -> str:
        return f"Please recommend the most popular book by {author}"

    with snapshot_test(snapshot) as snap:
        response = await recommend_book("Patrick Rothfuss")
        snap.set_response(response)

        book = response.parse()
        assert book.author.first_name == "Patrick"
        assert book.author.last_name == "Rothfuss"
        assert book.title == "THE NAME OF THE WIND"
        assert book.rating == 7


@pytest.mark.parametrize("model_id", STRUCTURED_OUTPUT_MODEL_IDS)
@pytest.mark.parametrize("formatting_mode", FORMATTING_MODES)
@pytest.mark.vcr
@pytest.mark.asyncio
async def test_structured_output_async_context(
    model_id: llm.ModelId,
    formatting_mode: llm.FormattingMode | None,
    snapshot: Snapshot,
) -> None:
    """Test asynchronous structured output with context."""

    format = (
        llm.format(Book, mode=formatting_mode) if formatting_mode is not None else Book
    )

    @llm.call(
        model_id,
        format=format,
    )
    async def recommend_book(ctx: llm.Context[str]) -> str:
        return f"Please recommend the most popular book by {ctx.deps}"

    ctx = llm.Context(deps="Patrick Rothfuss")
    with snapshot_test(snapshot) as snap:
        response = await recommend_book(ctx)
        snap.set_response(response)

        book = response.parse()
        assert book.author.first_name == "Patrick"
        assert book.author.last_name == "Rothfuss"
        assert book.title == "THE NAME OF THE WIND"
        assert book.rating == 7


# ============= STREAM TESTS =============


@pytest.mark.parametrize("model_id", STRUCTURED_OUTPUT_MODEL_IDS)
@pytest.mark.parametrize("formatting_mode", FORMATTING_MODES)
@pytest.mark.vcr
def test_structured_output_stream(
    model_id: llm.ModelId,
    formatting_mode: llm.FormattingMode | None,
    snapshot: Snapshot,
) -> None:
    """Test streaming structured output without context."""

    format = (
        llm.format(Book, mode=formatting_mode) if formatting_mode is not None else Book
    )

    @llm.call(
        model_id,
        format=format,
    )
    def recommend_book(author: str) -> str:
        return f"Please recommend the most popular book by {author}"

    with snapshot_test(snapshot) as snap:
        response = recommend_book.stream("Patrick Rothfuss")
        response.finish()

        snap.set_response(response)

        book = response.parse()
        assert book.author.first_name == "Patrick"
        assert book.author.last_name == "Rothfuss"
        assert book.title == "THE NAME OF THE WIND"
        assert book.rating == 7


@pytest.mark.parametrize("model_id", STRUCTURED_OUTPUT_MODEL_IDS)
@pytest.mark.parametrize("formatting_mode", FORMATTING_MODES)
@pytest.mark.vcr
def test_structured_output_stream_context(
    model_id: llm.ModelId,
    formatting_mode: llm.FormattingMode | None,
    snapshot: Snapshot,
) -> None:
    """Test streaming structured output with context."""

    format = (
        llm.format(Book, mode=formatting_mode) if formatting_mode is not None else Book
    )

    @llm.call(
        model_id,
        format=format,
    )
    def recommend_book(ctx: llm.Context[str]) -> str:
        return f"Please recommend the most popular book by {ctx.deps}"

    ctx = llm.Context(deps="Patrick Rothfuss")
    with snapshot_test(snapshot) as snap:
        response = recommend_book.stream(ctx)
        response.finish()

        snap.set_response(response)

        book = response.parse()
        assert book.author.first_name == "Patrick"
        assert book.author.last_name == "Rothfuss"
        assert book.title == "THE NAME OF THE WIND"
        assert book.rating == 7


# ============= ASYNC STREAM TESTS =============


@pytest.mark.parametrize("model_id", STRUCTURED_OUTPUT_MODEL_IDS)
@pytest.mark.parametrize("formatting_mode", FORMATTING_MODES)
@pytest.mark.vcr
@pytest.mark.asyncio
async def test_structured_output_async_stream(
    model_id: llm.ModelId,
    formatting_mode: llm.FormattingMode | None,
    snapshot: Snapshot,
) -> None:
    """Test async streaming structured output without context."""

    format = (
        llm.format(Book, mode=formatting_mode) if formatting_mode is not None else Book
    )

    @llm.call(
        model_id,
        format=format,
    )
    async def recommend_book(author: str) -> str:
        return f"Please recommend the most popular book by {author}"

    with snapshot_test(snapshot) as snap:
        response = await recommend_book.stream("Patrick Rothfuss")
        await response.finish()

        snap.set_response(response)

        book = response.parse()
        assert book.author.first_name == "Patrick"
        assert book.author.last_name == "Rothfuss"
        assert book.title == "THE NAME OF THE WIND"
        assert book.rating == 7


@pytest.mark.parametrize("model_id", STRUCTURED_OUTPUT_MODEL_IDS)
@pytest.mark.parametrize("formatting_mode", FORMATTING_MODES)
@pytest.mark.vcr
@pytest.mark.asyncio
async def test_structured_output_async_stream_context(
    model_id: llm.ModelId,
    formatting_mode: llm.FormattingMode | None,
    snapshot: Snapshot,
) -> None:
    """Test async streaming structured output with context."""

    format = (
        llm.format(Book, mode=formatting_mode) if formatting_mode is not None else Book
    )

    @llm.call(
        model_id,
        format=format,
    )
    async def recommend_book(ctx: llm.Context[str]) -> str:
        return f"Please recommend the most popular book by {ctx.deps}"

    ctx = llm.Context(deps="Patrick Rothfuss")
    with snapshot_test(snapshot) as snap:
        response = await recommend_book.stream(ctx)
        await response.finish()

        snap.set_response(response)

        book = response.parse()
        assert book.author.first_name == "Patrick"
        assert book.author.last_name == "Rothfuss"
        assert book.title == "THE NAME OF THE WIND"
        assert book.rating == 7



"""Tests for the Model class."""

import pytest

from mirascope import llm


def test_use_model_without_context() -> None:
    """Test that use_model creates a new Model when no context is set."""
    model = llm.use_model("openai/gpt-4o")

    assert model.provider_id == "openai"
    assert model.model_id == "openai/gpt-4o"


def test_use_model_with_context() -> None:
    """Test that use_model returns the context model when one is set."""
    with llm.model("anthropic/claude-sonnet-4-0"):
        model = llm.use_model("openai/gpt-4o")

        assert model.provider_id == "anthropic"
        assert model.model_id == "anthropic/claude-sonnet-4-0"


def test_use_model_as_context_manager() -> None:
    """Test that model can be used as a context manager directly."""
    with llm.Model("anthropic/claude-sonnet-4-0"):
        model = llm.use_model("openai/gpt-4o")

        assert model.provider_id == "anthropic"
        assert model.model_id == "anthropic/claude-sonnet-4-0"


def test_direct_model_instantiation_ignores_context() -> None:
    """Test that direct Model instantiation ignores context (hardcoding behavior)."""
    with llm.model("openai/claude-sonnet-4-0"):
        model = llm.Model("openai/gpt-4o")

        assert model.provider_id == "openai"
        assert model.model_id == "openai/gpt-4o"


def test_value_error_invalid_models() -> None:
    """Test that invalid model_ids raise appropriate errors."""

    # Test unregistered provider - need to actually call to trigger runtime error
    @llm.call("nonexistent/gpt-5-mini")
    def say_hi() -> str:
        return "hi"

    with pytest.raises(llm.NoRegisteredProviderError, match="No provider registered"):
        say_hi()

    # Test invalid model_id format (no slash)
    with pytest.raises(ValueError, match="Invalid model_id format"):
        llm.call("really-cool-model-i-heard-about")


def test_use_model_accepts_model_instance() -> None:
    """Test that use_model accepts a Model instance."""
    model_instance = llm.Model("anthropic/claude-sonnet-4-0", temperature=0.7)
    model = llm.use_model(model_instance)

    assert model.provider_id == "anthropic"
    assert model.model_id == "anthropic/claude-sonnet-4-0"
    assert model.params.get("temperature") == 0.7


def test_nested_context_manager_with_same_instance() -> None:
    """Test that the same Model instance can be used in nested contexts.

    Expected behavior: Each nested context should properly manage its own token
    and restore the previous state when exiting.
    """
    model = llm.Model("openai/gpt-4o")

    assert llm.model_from_context() is None

    with model:
        assert llm.model_from_context() is model

        with model:  # Inner context with SAME instance
            assert llm.model_from_context() is model

        assert llm.model_from_context() is model, (
            "Context should still be active after nested exit, but token was lost"
        )

    assert llm.model_from_context() is None


def test_provider_resolution_is_dynamic() -> None:
    """Test that Model dynamically resolves provider based on registry changes.

    This ensures that changes to the provider registry affect which provider
    is used by Model instances, even after the Model is created.
    """
    # Create a model - should auto-register and use default OpenAI provider
    model = llm.Model("openai/gpt-4o")
    initial_provider = model.provider
    assert initial_provider.id == "openai"

    # Register a custom provider for the same scope
    custom_provider = llm.register_provider("anthropic", scope="openai/")

    # The same Model instance should now resolve to the custom provider
    assert model.provider is custom_provider
    assert model.provider.id == "anthropic"

    # Register another provider for a more specific scope
    very_specific_provider = llm.register_provider("google", scope="openai/gpt-4o")

    # Should now use the most specific provider
    assert model.provider is very_specific_provider
    assert model.provider.id == "google"

    # A different model should use less specific provider
    other_model = llm.Model("openai/gpt-3.5-turbo")
    assert other_model.provider is custom_provider
    assert other_model.provider.id == "anthropic"



from mirascope import llm


@llm.call("openai/gpt-5-mini")
def recommend_book(genre: str):
    return f"Please recommend a book in {genre}."


# Override the model at runtime
with llm.model("anthropic/claude-sonnet-4-5", temperature=0.9):
    response = recommend_book("fantasy")
    print(response.pretty())



# This file was auto-generated by Fern from our API Definition.

import typing

OrganizationsUpdateSubscriptionRequestTargetPlan = typing.Union[
    typing.Literal["free", "pro", "team"], typing.Any
]



from __future__ import annotations

import abc
from collections.abc import Iterable, Sequence
from typing import TypeAlias

from mlx_lm.generate import GenerationResponse

from ....formatting import Format, FormatSpec, FormattableT
from ....messages import AssistantContent, Message
from ....responses import ChunkIterator
from ....tools import AnyToolSchema, BaseToolkit

TokenIds: TypeAlias = list[int]


class BaseEncoder(abc.ABC):
    """Abstract base class for Mirascope <> MLX encoding and decoding."""

    @abc.abstractmethod
    def encode_request(
        self,
        messages: Sequence[Message],
        tools: BaseToolkit[AnyToolSchema],
        format: FormatSpec[FormattableT] | None,
    ) -> tuple[Sequence[Message], Format[FormattableT] | None, TokenIds]:
        """Encode the request messages into a format suitable for the model.

        Args:
            messages: The sequence of messages to encode.
            tools: Optional sequence of tool schemas or toolkit for the model.
            format: Optional format specification for structured outputs.

        Returns:
            A tuple containing:
                - The processed messages
                - The format specification (if provided)
                - The encoded prompt as token IDs
        """

        ...

    @abc.abstractmethod
    def decode_response(
        self, stream: Iterable[GenerationResponse]
    ) -> tuple[AssistantContent, GenerationResponse | None]:
        """Decode a stream of MLX generation responses into assistant content.

        Args:
            stream: An iterable of MLX generation responses.

        Returns:
            A tuple containing:
                - The decoded assistant content
                - The final generation response (if available)
        """
        ...

    @abc.abstractmethod
    def decode_stream(self, stream: Iterable[GenerationResponse]) -> ChunkIterator:
        """Decode a stream of MLX generation responses into an iterable of chunks.

        Args:
            stream: An iterable of MLX generation responses.

        Returns:
            A ChunkIterator yielding content chunks for streaming responses.
        """
        ...



"""Tests for ContextRetryStreamResponse and AsyncContextRetryStreamResponse."""

import pytest

from mirascope import llm

from .conftest import CONNECTION_ERROR, RATE_LIMIT_ERROR, SERVER_ERROR, TIMEOUT_ERROR
from .mock_provider import MockProvider

# --- Sync ContextRetryStreamResponse tests ---


def test_context_stream_succeeds_first_attempt(mock_provider: MockProvider) -> None:
    """Test that a successful context stream completes without raising StreamRestarted."""
    ctx = llm.Context(deps=None)
    retry_model = llm.retry_model("mock/test-model", max_retries=2)

    response = retry_model.context_stream("Hello", ctx=ctx)

    # Consume the stream
    chunks = list(response.text_stream())

    assert "mock " in "".join(chunks)
    assert "response" in "".join(chunks)
    assert len(response.retry_failures) == 0
    assert mock_provider.stream_count == 1


def test_context_stream_raises_stream_restarted_on_error(
    mock_provider: MockProvider,
) -> None:
    """Test that StreamRestarted is raised when a retryable error occurs mid-stream."""
    mock_provider.set_stream_exceptions([CONNECTION_ERROR])
    ctx = llm.Context(deps=None)
    retry_model = llm.retry_model("mock/test-model", max_retries=2)

    response = retry_model.context_stream("Hello", ctx=ctx)

    # First iteration should raise StreamRestarted
    with pytest.raises(llm.StreamRestarted) as exc_info:
        list(response.text_stream())

    assert exc_info.value.failure.exception is CONNECTION_ERROR
    assert len(response.retry_failures) == 1
    assert [f.exception for f in response.retry_failures] == [CONNECTION_ERROR]


def test_context_stream_can_continue_after_restart(mock_provider: MockProvider) -> None:
    """Test that user can re-iterate after catching StreamRestarted."""
    mock_provider.set_stream_exceptions([CONNECTION_ERROR])
    ctx = llm.Context(deps=None)
    retry_model = llm.retry_model("mock/test-model", max_retries=2)

    response = retry_model.context_stream("Hello", ctx=ctx)
    restart_count = 0

    # Use the recommended pattern: catch StreamRestarted and re-iterate
    while True:
        try:
            chunks = list(response.text_stream())
            break
        except llm.StreamRestarted:
            restart_count += 1

    assert restart_count == 1
    assert "mock " in "".join(chunks)
    assert "response" in "".join(chunks)
    assert len(response.retry_failures) == 1
    assert mock_provider.stream_count == 2


def test_context_stream_raises_retries_exhausted_after_max_attempts(
    mock_provider: MockProvider,
) -> None:
    """Test that RetriesExhausted is raised when max attempts exhausted."""
    # Set more errors than max_attempts allows
    mock_provider.set_stream_exceptions([CONNECTION_ERROR, RATE_LIMIT_ERROR])
    ctx = llm.Context(deps=None)
    retry_model = llm.retry_model("mock/test-model", max_retries=1)

    response = retry_model.context_stream("Hello", ctx=ctx)

    # First attempt fails, catch StreamRestarted
    with pytest.raises(llm.StreamRestarted):
        list(response.text_stream())

    # Second attempt also fails, should raise RetriesExhausted (not StreamRestarted)
    with pytest.raises(llm.RetriesExhausted) as exc_info:
        list(response.text_stream())

    # RetriesExhausted contains both failures
    assert len(exc_info.value.failures) == 2
    assert isinstance(exc_info.value.failures[0].exception, llm.ConnectionError)
    assert isinstance(exc_info.value.failures[1].exception, llm.RateLimitError)
    assert mock_provider.stream_count == 2


def test_context_stream_model_property(mock_provider: MockProvider) -> None:  # noqa: ARG001
    """Test that model property returns RetryModel."""
    ctx = llm.Context(deps=None)
    retry_model = llm.retry_model("mock/test-model", max_retries=2)

    response = retry_model.context_stream("Hello", ctx=ctx)

    assert isinstance(response.model, llm.RetryModel)
    assert response.model.model_id == "mock/test-model"


def test_context_stream_delegated_properties(mock_provider: MockProvider) -> None:
    """Test that all properties delegate to the wrapped stream response."""
    ctx = llm.Context(deps=None)
    retry_model = llm.retry_model("mock/test-model", max_retries=1)

    response = retry_model.context_stream("Hello", ctx=ctx)

    # Consume the stream to populate properties
    list(response.chunk_stream())

    # Test all delegated properties
    assert response.raw_stream_events is not None
    assert response.chunks is not None
    assert response.provider_id == "mock"
    assert response.model_id == "mock/test-model"
    assert response.provider_model_name == "test-model"
    assert response.params is not None
    assert response.toolkit is not None
    assert response.messages is not None
    assert response.content is not None
    assert response.texts is not None
    assert response.tool_calls is not None
    assert response.thoughts is not None
    _ = response.finish_reason
    _ = response.usage
    assert response.consumed is True
    assert response.format is None
    assert response.parse() is None


def test_context_stream_execute_tools(mock_provider: MockProvider) -> None:  # noqa: ARG001
    """Test that execute_tools delegates to wrapped response with ctx."""
    ctx = llm.Context(deps=None)
    retry_model = llm.retry_model("mock/test-model", max_retries=1)

    response = retry_model.context_stream("Hello", ctx=ctx)
    response.finish()

    result = response.execute_tools(ctx)
    assert result == []


def test_context_stream_resume(mock_provider: MockProvider) -> None:
    """Test that resume returns a new ContextRetryStreamResponse."""
    ctx = llm.Context(deps=None)
    retry_model = llm.retry_model("mock/test-model", max_retries=2)

    response = retry_model.context_stream("Hello", ctx=ctx)
    response.finish()

    resumed = response.resume(ctx, "Follow up")

    assert isinstance(resumed, llm.ContextRetryStreamResponse)
    assert mock_provider.stream_count == 2


def test_context_stream_resume_uses_fallback_model(mock_provider: MockProvider) -> None:
    """Test that stream resume uses the fallback model when original succeeded on fallback."""
    # First stream: primary fails, fallback succeeds
    mock_provider.set_stream_exceptions([CONNECTION_ERROR])
    ctx = llm.Context(deps=None)
    retry_model = llm.retry_model(
        "mock/primary", max_retries=0, fallback_models=["mock/fallback"]
    )

    response = retry_model.context_stream("Hello", ctx=ctx)

    # First iteration fails (primary), raises StreamRestarted
    with pytest.raises(llm.StreamRestarted):
        list(response.chunk_stream())

    # After restart, should be on fallback model
    assert response.model.model_id == "mock/fallback"

    # Second iteration succeeds on fallback
    list(response.chunk_stream())

    # Resume should use the fallback model (which succeeded)
    resumed = response.resume(ctx, "Follow up")

    # Should start on fallback (no errors set)
    assert resumed.model.model_id == "mock/fallback"

    # Should succeed on fallback
    chunks = list(resumed.chunk_stream())
    assert len(chunks) > 0


def test_context_stream_resume_can_fall_back(mock_provider: MockProvider) -> None:
    """Test that stream resume can fall back to original model if fallback fails."""
    # First stream: primary fails, fallback succeeds
    mock_provider.set_stream_exceptions([CONNECTION_ERROR])
    ctx = llm.Context(deps=None)
    retry_model = llm.retry_model(
        "mock/primary", max_retries=0, fallback_models=["mock/fallback"]
    )

    response = retry_model.context_stream("Hello", ctx=ctx)

    # First iteration fails (primary), raises StreamRestarted
    with pytest.raises(llm.StreamRestarted):
        list(response.chunk_stream())

    # After restart, should be on fallback model
    assert response.model.model_id == "mock/fallback"

    # Second iteration succeeds on fallback
    list(response.chunk_stream())

    # For resume: fallback fails, primary succeeds
    mock_provider.set_stream_exceptions([RATE_LIMIT_ERROR])

    resumed = response.resume(ctx, "Follow up")

    # Should start on fallback model (which succeeded before)
    assert resumed.model.model_id == "mock/fallback"

    # First iteration of resume fails (fallback), raises StreamRestarted
    with pytest.raises(llm.StreamRestarted):
        list(resumed.chunk_stream())

    # After restart, should be on primary model
    assert resumed.model.model_id == "mock/primary"
    assert len(resumed.retry_failures) == 1
    assert resumed.retry_failures[0].model.model_id == "mock/fallback"

    # Second iteration succeeds on primary
    chunks = list(resumed.chunk_stream())
    assert len(chunks) > 0


# --- Async AsyncContextRetryStreamResponse tests ---


@pytest.mark.asyncio
async def test_async_context_stream_succeeds_first_attempt(
    mock_provider: MockProvider,
) -> None:
    """Test that a successful async context stream completes without raising StreamRestarted."""
    ctx = llm.Context(deps=None)
    retry_model = llm.retry_model("mock/test-model", max_retries=2)

    response = await retry_model.context_stream_async("Hello", ctx=ctx)

    # Consume the stream
    chunks = [chunk async for chunk in response.text_stream()]

    assert "mock " in "".join(chunks)
    assert "response" in "".join(chunks)
    assert len(response.retry_failures) == 0
    assert mock_provider.stream_count == 1


@pytest.mark.asyncio
async def test_async_context_stream_raises_stream_restarted_on_error(
    mock_provider: MockProvider,
) -> None:
    """Test that StreamRestarted is raised when a retryable error occurs mid-stream."""
    mock_provider.set_stream_exceptions([TIMEOUT_ERROR])
    ctx = llm.Context(deps=None)
    retry_model = llm.retry_model("mock/test-model", max_retries=2)

    response = await retry_model.context_stream_async("Hello", ctx=ctx)

    # First iteration should raise StreamRestarted
    with pytest.raises(llm.StreamRestarted) as exc_info:
        _ = [chunk async for chunk in response.text_stream()]

    assert exc_info.value.failure.exception is TIMEOUT_ERROR
    assert len(response.retry_failures) == 1


@pytest.mark.asyncio
async def test_async_context_stream_can_continue_after_restart(
    mock_provider: MockProvider,
) -> None:
    """Test that user can re-iterate after catching StreamRestarted."""
    mock_provider.set_stream_exceptions([SERVER_ERROR])
    ctx = llm.Context(deps=None)
    retry_model = llm.retry_model("mock/test-model", max_retries=2)

    response = await retry_model.context_stream_async("Hello", ctx=ctx)
    restart_count = 0

    # Use the recommended pattern: catch StreamRestarted and re-iterate
    while True:
        try:
            chunks = [chunk async for chunk in response.text_stream()]
            break
        except llm.StreamRestarted:
            restart_count += 1

    assert restart_count == 1
    assert "mock " in "".join(chunks)
    assert "response" in "".join(chunks)
    assert len(response.retry_failures) == 1
    assert mock_provider.stream_count == 2


@pytest.mark.asyncio
async def test_async_context_stream_raises_retries_exhausted_after_max_attempts(
    mock_provider: MockProvider,
) -> None:
    """Test that RetriesExhausted is raised when max attempts exhausted."""
    mock_provider.set_stream_exceptions([CONNECTION_ERROR, SERVER_ERROR])
    ctx = llm.Context(deps=None)
    retry_model = llm.retry_model("mock/test-model", max_retries=1)

    response = await retry_model.context_stream_async("Hello", ctx=ctx)

    # First attempt fails, catch StreamRestarted
    with pytest.raises(llm.StreamRestarted):
        _ = [chunk async for chunk in response.text_stream()]

    # Second attempt also fails, should raise RetriesExhausted
    with pytest.raises(llm.RetriesExhausted) as exc_info:
        _ = [chunk async for chunk in response.text_stream()]

    # RetriesExhausted contains both failures
    assert len(exc_info.value.failures) == 2
    assert isinstance(exc_info.value.failures[0].exception, llm.ConnectionError)
    assert isinstance(exc_info.value.failures[1].exception, llm.ServerError)
    assert mock_provider.stream_count == 2


@pytest.mark.asyncio
async def test_async_context_stream_model_property(
    mock_provider: MockProvider,
) -> None:  # noqa: ARG001
    """Test that model property returns RetryModel."""
    ctx = llm.Context(deps=None)
    retry_model = llm.retry_model("mock/test-model", max_retries=2)

    response = await retry_model.context_stream_async("Hello", ctx=ctx)

    assert isinstance(response.model, llm.RetryModel)
    assert response.model.model_id == "mock/test-model"


@pytest.mark.asyncio
async def test_async_context_stream_delegated_properties(
    mock_provider: MockProvider,
) -> None:
    """Test that all properties delegate to the wrapped async stream response."""
    ctx = llm.Context(deps=None)
    retry_model = llm.retry_model("mock/test-model", max_retries=1)

    response = await retry_model.context_stream_async("Hello", ctx=ctx)

    # Consume the stream to populate properties
    async for _ in response.chunk_stream():
        pass

    # Test all delegated properties
    assert response.raw_stream_events is not None
    assert response.chunks is not None
    assert response.provider_id == "mock"
    assert response.model_id == "mock/test-model"
    assert response.provider_model_name == "test-model"
    assert response.params is not None
    assert response.toolkit is not None
    assert response.messages is not None
    assert response.content is not None
    assert response.texts is not None
    assert response.tool_calls is not None
    assert response.thoughts is not None
    _ = response.finish_reason
    _ = response.usage
    assert response.consumed is True
    assert response.format is None
    assert response.parse() is None


@pytest.mark.asyncio
async def test_async_context_stream_execute_tools(
    mock_provider: MockProvider,
) -> None:  # noqa: ARG001
    """Test that async execute_tools delegates to wrapped response with ctx."""
    ctx = llm.Context(deps=None)
    retry_model = llm.retry_model("mock/test-model", max_retries=1)

    response = await retry_model.context_stream_async("Hello", ctx=ctx)
    await response.finish()

    result = await response.execute_tools(ctx)
    assert result == []


@pytest.mark.asyncio
async def test_async_context_stream_resume(mock_provider: MockProvider) -> None:
    """Test that async resume returns a new AsyncContextRetryStreamResponse."""
    ctx = llm.Context(deps=None)
    retry_model = llm.retry_model("mock/test-model", max_retries=2)

    response = await retry_model.context_stream_async("Hello", ctx=ctx)
    await response.finish()

    resumed = await response.resume(ctx, "Follow up")

    assert isinstance(resumed, llm.AsyncContextRetryStreamResponse)
    assert mock_provider.stream_count == 2


@pytest.mark.asyncio
async def test_async_context_stream_resume_uses_fallback_model(
    mock_provider: MockProvider,
) -> None:
    """Test that async stream resume uses the fallback model when original succeeded on fallback."""
    # First stream: primary fails, fallback succeeds
    mock_provider.set_stream_exceptions([CONNECTION_ERROR])
    ctx = llm.Context(deps=None)
    retry_model = llm.retry_model(
        "mock/primary", max_retries=0, fallback_models=["mock/fallback"]
    )

    response = await retry_model.context_stream_async("Hello", ctx=ctx)

    # First iteration fails (primary), raises StreamRestarted
    with pytest.raises(llm.StreamRestarted):
        _ = [chunk async for chunk in response.chunk_stream()]

    # After restart, should be on fallback model
    assert response.model.model_id == "mock/fallback"

    # Second iteration succeeds on fallback
    _ = [chunk async for chunk in response.chunk_stream()]

    # Resume should use the fallback model (which succeeded)
    resumed = await response.resume(ctx, "Follow up")

    # Should start on fallback (no errors set)
    assert resumed.model.model_id == "mock/fallback"

    # Should succeed on fallback
    chunks = [chunk async for chunk in resumed.chunk_stream()]
    assert len(chunks) > 0


@pytest.mark.asyncio
async def test_async_context_stream_resume_can_fall_back(
    mock_provider: MockProvider,
) -> None:
    """Test that async stream resume can fall back to original model if fallback fails."""
    # First stream: primary fails, fallback succeeds
    mock_provider.set_stream_exceptions([CONNECTION_ERROR])
    ctx = llm.Context(deps=None)
    retry_model = llm.retry_model(
        "mock/primary", max_retries=0, fallback_models=["mock/fallback"]
    )

    response = await retry_model.context_stream_async("Hello", ctx=ctx)

    # First iteration fails (primary), raises StreamRestarted
    with pytest.raises(llm.StreamRestarted):
        _ = [chunk async for chunk in response.chunk_stream()]

    # After restart, should be on fallback model
    assert response.model.model_id == "mock/fallback"

    # Second iteration succeeds on fallback
    _ = [chunk async for chunk in response.chunk_stream()]

    # For resume: fallback fails, primary succeeds
    mock_provider.set_stream_exceptions([RATE_LIMIT_ERROR])

    resumed = await response.resume(ctx, "Follow up")

    # Should start on fallback model (which succeeded before)
    assert resumed.model.model_id == "mock/fallback"

    # First iteration of resume fails (fallback), raises StreamRestarted
    with pytest.raises(llm.StreamRestarted):
        _ = [chunk async for chunk in resumed.chunk_stream()]

    # After restart, should be on primary model
    assert resumed.model.model_id == "mock/primary"
    assert len(resumed.retry_failures) == 1
    assert resumed.retry_failures[0].model.model_id == "mock/fallback"

    # Second iteration succeeds on primary
    chunks = [chunk async for chunk in resumed.chunk_stream()]
    assert len(chunks) > 0



from inline_snapshot import snapshot

from mirascope.llm import (
    AssistantMessage,
    Text,
    UserMessage,
)

sync_snapshot = snapshot(
    {
        "response": {
            "provider_id": "openai",
            "model_id": "openai/gpt-5:responses",
            "provider_model_name": "gpt-5:responses",
            "params": {"thinking": {"level": "medium", "include_thoughts": False}},
            "finish_reason": None,
            "usage": {
                "input_tokens": 32,
                "output_tokens": 1159,
                "cache_read_tokens": 0,
                "cache_write_tokens": 0,
                "reasoning_tokens": 1152,
                "provider_tool_usage": None,
                "raw": "ResponseUsage(input_tokens=32, input_tokens_details=InputTokensDetails(cached_tokens=0), output_tokens=1159, output_tokens_details=OutputTokensDetails(reasoning_tokens=1152), total_tokens=1191)",
                "total_tokens": 1191,
            },
            "messages": [
                UserMessage(
                    content=[
                        Text(
                            text="How many primes below 400 contain 79 as a substring? Answer ONLY with the number, not sharing which primes they are."
                        )
                    ]
                ),
                AssistantMessage(
                    content=[Text(text="3")],
                    provider_id="openai",
                    model_id="openai/gpt-5:responses",
                    provider_model_name="gpt-5:responses",
                    raw_message=[
                        {
                            "id": "rs_085d300305cb9e0f00696aefb6042481968e79bd97da8920f0",
                            "summary": [],
                            "type": "reasoning",
                        },
                        {
                            "id": "msg_085d300305cb9e0f00696aefc6a27081969e5e73c231e5be1d",
                            "content": [
                                {
                                    "annotations": [],
                                    "text": "3",
                                    "type": "output_text",
                                    "logprobs": [],
                                }
                            ],
                            "role": "assistant",
                            "status": "completed",
                            "type": "message",
                        },
                    ],
                ),
            ],
            "format": None,
            "tools": [],
        }
    }
)
async_snapshot = snapshot(
    {
        "response": {
            "provider_id": "openai",
            "model_id": "openai/gpt-5:responses",
            "provider_model_name": "gpt-5:responses",
            "params": {"thinking": {"level": "medium", "include_thoughts": False}},
            "finish_reason": None,
            "usage": {
                "input_tokens": 32,
                "output_tokens": 1287,
                "cache_read_tokens": 0,
                "cache_write_tokens": 0,
                "reasoning_tokens": 1280,
                "provider_tool_usage": None,
                "raw": "ResponseUsage(input_tokens=32, input_tokens_details=InputTokensDetails(cached_tokens=0), output_tokens=1287, output_tokens_details=OutputTokensDetails(reasoning_tokens=1280), total_tokens=1319)",
                "total_tokens": 1319,
            },
            "messages": [
                UserMessage(
                    content=[
                        Text(
                            text="How many primes below 400 contain 79 as a substring? Answer ONLY with the number, not sharing which primes they are."
                        )
                    ]
                ),
                AssistantMessage(
                    content=[Text(text="3")],
                    provider_id="openai",
                    model_id="openai/gpt-5:responses",
                    provider_model_name="gpt-5:responses",
                    raw_message=[
                        {
                            "id": "rs_05dbc5316d1fcc3100696af01c615c81969031391e23c74b37",
                            "summary": [],
                            "type": "reasoning",
                        },
                        {
                            "id": "msg_05dbc5316d1fcc3100696af02c68fc8196b01531821357a5ba",
                            "content": [
                                {
                                    "annotations": [],
                                    "text": "3",
                                    "type": "output_text",
                                    "logprobs": [],
                                }
                            ],
                            "role": "assistant",
                            "status": "completed",
                            "type": "message",
                        },
                    ],
                ),
            ],
            "format": None,
            "tools": [],
        }
    }
)
stream_snapshot = snapshot(
    {
        "response": {
            "provider_id": "openai",
            "model_id": "openai/gpt-5:responses",
            "provider_model_name": "gpt-5:responses",
            "finish_reason": None,
            "messages": [
                UserMessage(
                    content=[
                        Text(
                            text="How many primes below 400 contain 79 as a substring? Answer ONLY with the number, not sharing which primes they are."
                        )
                    ]
                ),
                AssistantMessage(
                    content=[Text(text="3")],
                    provider_id="openai",
                    model_id="openai/gpt-5:responses",
                    provider_model_name="gpt-5:responses",
                    raw_message=[
                        {
                            "id": "rs_00be767a684ac49d00696aefefbda881949385c2abc3a0a1c0",
                            "summary": [],
                            "type": "reasoning",
                        },
                        {
                            "id": "msg_00be767a684ac49d00696aeff84d348194900f2557198e218d",
                            "content": [
                                {
                                    "annotations": [],
                                    "text": "3",
                                    "type": "output_text",
                                    "logprobs": [],
                                }
                            ],
                            "role": "assistant",
                            "status": "completed",
                            "type": "message",
                        },
                    ],
                ),
            ],
            "format": None,
            "tools": [],
            "usage": {
                "input_tokens": 32,
                "output_tokens": 647,
                "cache_read_tokens": 0,
                "cache_write_tokens": 0,
                "reasoning_tokens": 640,
                "provider_tool_usage": None,
                "raw": "None",
                "total_tokens": 679,
            },
            "n_chunks": 3,
        }
    }
)
async_stream_snapshot = snapshot(
    {
        "response": {
            "provider_id": "openai",
            "model_id": "openai/gpt-5:responses",
            "provider_model_name": "gpt-5:responses",
            "finish_reason": None,
            "messages": [
                UserMessage(
                    content=[
                        Text(
                            text="How many primes below 400 contain 79 as a substring? Answer ONLY with the number, not sharing which primes they are."
                        )
                    ]
                ),
                AssistantMessage(
                    content=[Text(text="3")],
                    provider_id="openai",
                    model_id="openai/gpt-5:responses",
                    provider_model_name="gpt-5:responses",
                    raw_message=[
                        {
                            "id": "rs_0b3fa1fc0060d42b00696af059c1748197b3a1e949827e434c",
                            "summary": [],
                            "type": "reasoning",
                        },
                        {
                            "id": "msg_0b3fa1fc0060d42b00696af06545c48197afeccc2f4dae865a",
                            "content": [
                                {
                                    "annotations": [],
                                    "text": "3",
                                    "type": "output_text",
                                    "logprobs": [],
                                }
                            ],
                            "role": "assistant",
                            "status": "completed",
                            "type": "message",
                        },
                    ],
                ),
            ],
            "format": None,
            "tools": [],
            "usage": {
                "input_tokens": 32,
                "output_tokens": 839,
                "cache_read_tokens": 0,
                "cache_write_tokens": 0,
                "reasoning_tokens": 832,
                "provider_tool_usage": None,
                "raw": "None",
                "total_tokens": 871,
            },
            "n_chunks": 3,
        }
    }
)



"""Type definitions for the two-phase export system.

This module defines the event types and data structures used for
immediate start event transmission and batched end event export.
"""

from enum import Enum
from typing import Literal, TypedDict

from opentelemetry.util.types import AttributeValue

# TODO: unify/DRY types in the _internal package

SpanKind = Literal["CLIENT", "SERVER", "PRODUCER", "CONSUMER", "INTERNAL"]
StatusCode = Literal["UNSET", "OK", "ERROR"]


class SpanEventType(str, Enum):
    """Event types for span lifecycle tracking."""

    SPAN_STARTED = "span_started"
    SPAN_UPDATED = "span_updated"
    SPAN_COMPLETED = "span_completed"


class SpanEvent(TypedDict):
    """Individual event within a span."""

    name: str
    """The name of the event."""

    timestamp: int
    """Nanoseconds since epoch (OTel standard)."""

    attributes: dict[str, AttributeValue]
    """Event-specific attributes."""


class Status(TypedDict):
    """Status representation for serialization."""

    code: StatusCode
    """The status code: UNSET, OK, or ERROR."""

    description: str | None
    """Optional human-readable description of the status."""


class SpanContextDict(TypedDict):
    """Span context for links and parent references."""

    trace_id: str
    """The trace ID as a 32-character hex string."""

    span_id: str
    """The span ID as a 16-character hex string."""

    trace_flags: int
    """Trace flags (e.g., for sampling decisions)."""

    trace_state: str | None
    """Optional vendor-specific trace state."""


class Link(TypedDict):
    """Link representation for span relationships."""

    context: SpanContextDict
    """The linked span's context."""

    attributes: dict[str, AttributeValue]
    """Attributes describing the link."""


class SpanStartEvent(TypedDict):
    """Minimal span data for immediate transmission.

    This event is sent immediately when a span starts to provide
    real-time visibility into long-running operations.
    """

    trace_id: str
    """The trace ID as a 32-character hex string."""

    span_id: str
    """The span ID as a 16-character hex string."""

    parent_span_id: str | None
    """The parent span ID if this is a child span."""

    name: str
    """The name of the span."""

    start_time: int
    """Nanoseconds since epoch."""

    kind: SpanKind
    """The span kind: CLIENT, SERVER, PRODUCER, CONSUMER, or INTERNAL."""

    attributes: dict[str, AttributeValue]
    """Minimal required attributes only."""


class SpanUpdateEvent(TypedDict, total=False):
    """Incremental updates to span data.

    These events are batched and sent periodically to update
    span attributes or add events without waiting for completion.
    """

    trace_id: str
    """The trace ID as a 32-character hex string."""

    span_id: str
    """The span ID as a 16-character hex string."""

    timestamp: int
    """Nanoseconds since epoch."""

    attributes: dict[str, AttributeValue]
    """Additional or updated attributes."""

    events: list[SpanEvent]
    """New events to add to the span."""


class SpanCompleteEvent(TypedDict):
    """Complete span data for batch export.

    This event contains all span data and is sent when the span
    completes, typically in batches for efficiency.
    """

    trace_id: str
    """The trace ID as a 32-character hex string."""

    span_id: str
    """The span ID as a 16-character hex string."""

    parent_span_id: str | None
    """The parent span ID if this is a child span."""

    name: str
    """The name of the span."""

    kind: SpanKind
    """The span kind: CLIENT, SERVER, PRODUCER, CONSUMER, or INTERNAL."""

    start_time: int
    """Nanoseconds since epoch."""

    end_time: int
    """Nanoseconds since epoch."""

    status: Status
    """The final status of the span."""

    attributes: dict[str, AttributeValue]
    """All span attributes."""

    events: list[SpanEvent]
    """All events that occurred during the span."""

    links: list[Link]
    """Links to other spans."""



"""Stream classes for streaming assistant content parts."""

from abc import ABC, abstractmethod
from collections.abc import AsyncIterator, Iterator
from typing import Generic, Literal, TypeAlias, TypeVar

from ..content import (
    AssistantContentPart,
    Text,
    TextChunk,
    Thought,
    ThoughtChunk,
    ToolCall,
    ToolCallChunk,
)

ContentT = TypeVar("ContentT", bound=AssistantContentPart)
"""Type variable for content types (Text, Thought, ToolCall)."""

DeltaT = TypeVar("DeltaT")
"""Type variable for the deltas that the stream provides on iteration."""


class BaseStream(ABC, Generic[ContentT, DeltaT]):
    """Base class for synchronous streaming content."""

    @abstractmethod
    def __iter__(self) -> Iterator[DeltaT]:
        """Iterate over content deltas as they arrive."""
        ...

    @abstractmethod
    def collect(self) -> ContentT:
        """Collect all chunks and return the final content."""
        ...


class BaseAsyncStream(ABC, Generic[ContentT, DeltaT]):
    """Base class for asynchronous streaming content."""

    @abstractmethod
    def __aiter__(self) -> AsyncIterator[DeltaT]:
        """Asynchronously iterate over content deltas as they arrive."""
        ...

    @abstractmethod
    async def collect(self) -> ContentT:
        """Asynchronously collect all chunks and return the final content."""
        ...


class TextStream(BaseStream[Text, str]):
    """Synchronous text stream implementation."""

    type: Literal["text_stream"] = "text_stream"

    content_type: Literal["text"] = "text"
    """The type of content stored in this stream."""

    partial_text: str
    """The accumulated text content as chunks are received."""

    def __init__(
        self,
        chunk_iterator: Iterator[TextChunk],
    ) -> None:
        """Initialize TextStream with a start chunk and chunk iterator.

        Args:
            chunk_iterator: Iterator providing subsequent chunks.
        """
        self.partial_text = ""
        self._chunk_iterator = chunk_iterator

    def __iter__(self) -> Iterator[str]:
        """Iterate over text deltas as they are received.

        Yields:
            str: Each delta string containing text.
        """
        for chunk in self._chunk_iterator:
            delta = chunk.delta
            self.partial_text += delta
            yield delta

    def collect(self) -> Text:
        """Collect all chunks and return the final Text content.

        Returns:
            Text: The complete text content after consuming all chunks.
        """
        for _ in self:
            pass
        return Text(text=self.partial_text)


class AsyncTextStream(BaseAsyncStream[Text, str]):
    """Asynchronous text stream implementation."""

    type: Literal["async_text_stream"] = "async_text_stream"

    content_type: Literal["text"] = "text"
    """The type of content stored in this stream."""

    partial_text: str
    """The accumulated text content as chunks are received."""

    def __init__(
        self,
        chunk_iterator: AsyncIterator[TextChunk],
    ) -> None:
        """Initialize AsyncTextStream with a chunk iterator.

        Args:
            chunk_iterator: AsyncIterator providing subsequent chunks.
        """
        self.partial_text = ""
        self._chunk_iterator = chunk_iterator

    async def __aiter__(self) -> AsyncIterator[str]:
        """Asynchronously iterate over text deltas as they are received.

        Yields:
            str: Each delta string containing text.
        """
        async for chunk in self._chunk_iterator:
            delta = chunk.delta
            self.partial_text += delta
            yield delta

    async def collect(self) -> Text:
        """Asynchronously collect all chunks and return the final Text content.

        Returns:
            Text: The complete text content after consuming all chunks.
        """
        async for _ in self:
            pass
        return Text(text=self.partial_text)


class ThoughtStream(BaseStream[Thought, str]):
    """Synchronous thought stream implementation."""

    type: Literal["thought_stream"] = "thought_stream"

    content_type: Literal["thought"] = "thought"
    """The type of content stored in this stream."""

    partial_thought: str
    """The accumulated thought content as chunks are received."""

    def __init__(
        self,
        chunk_iterator: Iterator[ThoughtChunk],
    ) -> None:
        """Initialize ThoughtStream with a chunk iterator.

        Args:
            chunk_iterator: Iterator providing subsequent chunks.
        """
        self.partial_thought = ""
        self._chunk_iterator = chunk_iterator

    def __iter__(self) -> Iterator[str]:
        """Iterate over thought deltas as they are received.

        Yields:
            str: Each delta string containing thought text.
        """
        for chunk in self._chunk_iterator:
            delta = chunk.delta
            self.partial_thought += delta
            yield delta

    def collect(self) -> Thought:
        """Collect all chunks and return the final Thought content.

        Returns:
            Thought: The complete thought content after consuming all chunks.
        """
        for _ in self:
            pass
        return Thought(thought=self.partial_thought)


class AsyncThoughtStream(BaseAsyncStream[Thought, str]):
    """Asynchronous thought stream implementation."""

    type: Literal["async_thought_stream"] = "async_thought_stream"

    content_type: Literal["thought"] = "thought"
    """The type of content stored in this stream."""

    partial_thought: str
    """The accumulated thought content as chunks are received."""

    def __init__(
        self,
        chunk_iterator: AsyncIterator[ThoughtChunk],
    ) -> None:
        """Initialize AsyncThoughtStream with a chunk iterator.

        Args:
            chunk_iterator: AsyncIterator providing subsequent chunks.
        """
        self.partial_thought = ""
        self._chunk_iterator = chunk_iterator

    async def __aiter__(self) -> AsyncIterator[str]:
        """Asynchronously iterate over thought deltas as they are received.

        Yields:
            str: Each delta string containing thought text.
        """
        async for chunk in self._chunk_iterator:
            delta = chunk.delta
            self.partial_thought += delta
            yield delta

    async def collect(self) -> Thought:
        """Asynchronously collect all chunks and return the final Thought content.

        Returns:
            Thought: The complete thought content after consuming all chunks.
        """
        async for _ in self:
            pass
        return Thought(thought=self.partial_thought)


class ToolCallStream(BaseStream[ToolCall, str]):
    """Synchronous tool call stream implementation."""

    type: Literal["tool_call_stream"] = "tool_call_stream"

    content_type: Literal["tool_call"] = "tool_call"
    """The type of content stored in this stream."""

    tool_id: str
    """A unique identifier for this tool call."""

    tool_name: str
    """The name of the tool being called."""

    partial_args: str
    """The accumulated tool arguments as chunks are received."""

    def __init__(
        self,
        tool_id: str,
        tool_name: str,
        chunk_iterator: Iterator[ToolCallChunk],
    ) -> None:
        """Initialize ToolCallStream with tool metadata and chunk iterator.

        Args:
            tool_id: A unique identifier for this tool call.
            tool_name: The name of the tool being called.
            chunk_iterator: Iterator providing subsequent chunks.
        """
        self.tool_id = tool_id
        self.tool_name = tool_name
        self.partial_args = ""
        self._chunk_iterator = chunk_iterator

    def __iter__(self) -> Iterator[str]:
        """Iterate over tool call argument deltas as they are received.

        Yields:
            str: Each delta string containing JSON argument fragments.
        """
        for chunk in self._chunk_iterator:
            delta = chunk.delta
            self.partial_args += delta
            yield delta

    def collect(self) -> ToolCall:
        """Collect all chunks and return the final ToolCall content.

        Returns:
            ToolCall: The complete tool call after consuming all chunks.
        """
        for _ in self:
            pass
        return ToolCall(id=self.tool_id, name=self.tool_name, args=self.partial_args)


class AsyncToolCallStream(BaseAsyncStream[ToolCall, str]):
    """Asynchronous tool call stream implementation."""

    type: Literal["async_tool_call_stream"] = "async_tool_call_stream"

    content_type: Literal["tool_call"] = "tool_call"
    """The type of content stored in this stream."""

    tool_id: str
    """A unique identifier for this tool call."""

    tool_name: str
    """The name of the tool being called."""

    partial_args: str
    """The accumulated tool arguments as chunks are received."""

    def __init__(
        self,
        tool_id: str,
        tool_name: str,
        chunk_iterator: AsyncIterator[ToolCallChunk],
    ) -> None:
        """Initialize AsyncToolCallStream with tool metadata and chunk iterator.

        Args:
            tool_id: A unique identifier for this tool call.
            tool_name: The name of the tool being called.
            chunk_iterator: AsyncIterator providing subsequent chunks.
        """
        self.tool_id = tool_id
        self.tool_name = tool_name
        self.partial_args = ""
        self._chunk_iterator = chunk_iterator

    async def __aiter__(self) -> AsyncIterator[str]:
        """Asynchronously iterate over tool call argument deltas as they are received.

        Yields:
            str: Each delta string containing JSON argument fragments.
        """
        async for chunk in self._chunk_iterator:
            delta = chunk.delta
            self.partial_args += delta
            yield delta

    async def collect(self) -> ToolCall:
        """Asynchronously collect all chunks and return the final ToolCall content.

        Returns:
            ToolCall: The complete tool call after consuming all chunks.
        """
        async for _ in self:
            pass
        return ToolCall(id=self.tool_id, name=self.tool_name, args=self.partial_args)


Stream: TypeAlias = TextStream | ToolCallStream | ThoughtStream
"""A synchronous assistant content stream."""

AsyncStream: TypeAlias = AsyncTextStream | AsyncToolCallStream | AsyncThoughtStream
"""An asynchronous assistant content stream."""

__all__ = [
    "AsyncStream",
    "AsyncTextStream",
    "AsyncThoughtStream",
    "AsyncToolCallStream",
    "BaseAsyncStream",
    "BaseStream",
    "Stream",
    "TextStream",
    "ThoughtStream",
    "ToolCallStream",
]



from mirascope import llm

model = llm.model("openai/gpt-5")
response = model.call("Tell me a joke.")

# response.content contains all content parts: Text, ToolCall, Thought
for part in response.content:
    print(f"{type(part).__name__}: {part}")

# Filtered accessors for specific content types
for text in response.texts:
    print(f"Text: {text.text}")

for thought in response.thoughts:
    print(f"Thought: {thought.thought}")

for tool_call in response.tool_calls:
    print(f"Tool call: {tool_call.name}({tool_call.args})")



from mirascope import llm


@llm.prompt
def recommend_book(genre: str):
    return f"Please recommend a book in {genre}."


# Use llm.Model when you need to configure parameters
model = llm.Model("openai/gpt-4o", temperature=0.9)
response = recommend_book(model, "fantasy")
print(response.pretty())



"""Simple smoke tests for the prompt decorator."""

import pytest

from mirascope import llm


class TestPromptDecoratorSmokeTests:
    """Smoke tests for each prompt type.

    This just checks that we're returning the correct Prompt type as a general
    smoke test. We don't need to unit test tools, format, etc, because the e2e
    tests make extensive use of the `llm.prompt` decorator (via `llm.call`),
    so we get full coverage of the relevant code.
    """

    def test_decorator_construction(self) -> None:
        decorator: llm.PromptDecorator[llm.Tool, None] = llm.prompt()
        assert isinstance(decorator, llm.PromptDecorator)

    def test_sync_prompt_creation(self) -> None:
        """Test that a sync function creates a Prompt."""

        @llm.prompt
        def my_prompt(question: str) -> str:
            return f"Answer this: {question}"

        assert isinstance(my_prompt, llm.Prompt)

    @pytest.mark.asyncio
    async def test_async_prompt_creation(self) -> None:
        """Test that an async function creates an AsyncPrompt."""

        @llm.prompt
        async def my_async_prompt(question: str) -> str:
            return f"Answer this: {question}"

        assert isinstance(my_async_prompt, llm.AsyncPrompt)

    def test_context_prompt_creation(self) -> None:
        """Test that a sync function with ctx parameter creates a ContextPrompt."""

        @llm.prompt
        def my_context_prompt(ctx: llm.Context[str], question: str) -> str:
            return f"Context: {ctx.deps}. Question: {question}"

        assert isinstance(my_context_prompt, llm.ContextPrompt)

    @pytest.mark.asyncio
    async def test_async_context_prompt_creation(self) -> None:
        """Test that an async function with ctx parameter creates an AsyncContextPrompt."""

        @llm.prompt
        async def my_async_context_prompt(ctx: llm.Context[str], question: str) -> str:
            return f"Context: {ctx.deps}. Question: {question}"

        assert isinstance(my_async_context_prompt, llm.AsyncContextPrompt)



from mirascope import llm


@llm.call("openai/gpt-4o-mini", format=dict[str, str])
def recommend_books(genres: list[str]):
    return f"Recommend a book for each of the following genres: {', '.join(genres)}"


recommendations = recommend_books(["scifi", "fantasy", "romantasy"]).parse()
print(recommendations)
# {'scifi': 'Dune', 'fantasy': 'The Name of the Wind', 'romantasy': 'A Court of Thorns and Roses'}



from inline_snapshot import snapshot

from mirascope.llm import (
    AssistantMessage,
    Text,
    ToolCall,
    ToolExecutionError,
    ToolOutput,
    UserMessage,
)

test_snapshot = snapshot(
    {
        "response": {
            "provider_id": "openai",
            "model_id": "openai/gpt-4o:completions",
            "provider_model_name": "gpt-4o:completions",
            "params": {},
            "finish_reason": None,
            "usage": {
                "input_tokens": 113,
                "output_tokens": 32,
                "cache_read_tokens": 0,
                "cache_write_tokens": 0,
                "reasoning_tokens": 0,
                "provider_tool_usage": None,
                "raw": "CompletionUsage(completion_tokens=32, prompt_tokens=113, total_tokens=145, completion_tokens_details=CompletionTokensDetails(accepted_prediction_tokens=0, audio_tokens=0, reasoning_tokens=0, rejected_prediction_tokens=0), prompt_tokens_details=PromptTokensDetails(audio_tokens=0, cached_tokens=0))",
                "total_tokens": 145,
            },
            "messages": [
                UserMessage(
                    content=[
                        Text(
                            text="Use the test tool to retrieve the secret phrase. The passphrase is 'portal"
                        )
                    ]
                ),
                AssistantMessage(
                    content=[
                        ToolCall(
                            id="call_Usk6zffuhvy2YSSFmCyLHBlv",
                            name="passphrase_test_tool",
                            args='{"passphrase":"portal"}',
                        )
                    ],
                    provider_id="openai",
                    model_id="openai/gpt-4o:completions",
                    provider_model_name="gpt-4o:completions",
                    raw_message={
                        "role": "assistant",
                        "annotations": [],
                        "tool_calls": [
                            {
                                "id": "call_Usk6zffuhvy2YSSFmCyLHBlv",
                                "function": {
                                    "arguments": '{"passphrase":"portal"}',
                                    "name": "passphrase_test_tool",
                                },
                                "type": "function",
                            }
                        ],
                    },
                ),
                UserMessage(
                    content=[
                        ToolOutput(
                            id="call_Usk6zffuhvy2YSSFmCyLHBlv",
                            name="passphrase_test_tool",
                            result="Incorrect passhrase: The correct passphrase is 'cake'. Try again.",
                            error=ToolExecutionError(
                                "Incorrect passhrase: The correct passphrase is 'cake'. Try again."
                            ),
                        )
                    ]
                ),
                AssistantMessage(
                    content=[
                        Text(
                            text="It seems the passphrase you provided was incorrect. The correct passphrase is 'cake'. Would you like me to try again with the correct passphrase?"
                        )
                    ],
                    provider_id="openai",
                    model_id="openai/gpt-4o:completions",
                    provider_model_name="gpt-4o:completions",
                    raw_message={
                        "content": "It seems the passphrase you provided was incorrect. The correct passphrase is 'cake'. Would you like me to try again with the correct passphrase?",
                        "role": "assistant",
                        "annotations": [],
                    },
                ),
            ],
            "format": None,
            "tools": [
                {
                    "name": "passphrase_test_tool",
                    "description": "A tool that must be called with a passphrase.",
                    "parameters": """\
{
  "properties": {
    "passphrase": {
      "title": "Passphrase",
      "type": "string"
    }
  },
  "required": [
    "passphrase"
  ],
  "additionalProperties": false,
  "defs": null
}\
""",
                    "strict": None,
                }
            ],
        },
        "tool_outputs": [
            [
                ToolOutput(
                    id="call_Usk6zffuhvy2YSSFmCyLHBlv",
                    name="passphrase_test_tool",
                    result="Incorrect passhrase: The correct passphrase is 'cake'. Try again.",
                    error=ToolExecutionError(
                        "Incorrect passhrase: The correct passphrase is 'cake'. Try again."
                    ),
                )
            ]
        ],
    }
)



from pydantic import BaseModel

from mirascope import llm


class Book(BaseModel):
    title: str
    author: str


@llm.call("openai/gpt-5-mini", format=llm.format(Book, mode="strict"))
def recommend_book(genre: str):
    return f"Recommend a {genre} book."


response = recommend_book("fantasy")
book = response.parse()
print(f"{book.title} by {book.author}")



from mirascope import llm

# Messages can contain multiple content pieces of different types
image = llm.Image.from_url("https://example.com/chart.png")

message = llm.messages.user(
    [
        "Here's the sales chart from Q3.",
        image,
        "Can you summarize the trends?",
    ]
)



"""OpenTelemetry Gen AI Semantic Conventions types."""

from __future__ import annotations

from .gen_ai_input_messages import ChatMessage, InputMessages
from .gen_ai_output_messages import FinishReason, OutputMessage, OutputMessages
from .gen_ai_system_instructions import SystemInstructions
from .shared import (
    BlobPart,
    FilePart,
    GenericPart,
    Modality,
    ReasoningPart,
    Role,
    TextPart,
    ToolCallRequestPart,
    ToolCallResponsePart,
    UriPart,
)

__all__ = [
    "BlobPart",
    "ChatMessage",
    "FilePart",
    "FinishReason",
    "GenericPart",
    "InputMessages",
    "Modality",
    "OutputMessage",
    "OutputMessages",
    "ReasoningPart",
    "Role",
    "SystemInstructions",
    "TextPart",
    "ToolCallRequestPart",
    "ToolCallResponsePart",
    "UriPart",
]
